
#define VERSION "2011.04.30"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <string>

#include <list>
#include <vector>

#ifdef WIN32
	#include <process.h>
	#include <windows.h>
	typedef HANDLE serial_handle_t;
	#define DEFAULT_COM_PORT "\\\\.\\COM1"
	#pragma warning(disable:4996)
#else
	#include <unistd.h>
	#include <fcntl.h>
	#include <termios.h>
	#include <sys/stat.h>
	typedef int serial_handle_t;
	#ifdef Linux
		#define DEFAULT_COM_PORT "/dev/ttyS0"
	#endif
	#ifdef FreeBSD
		#define DEFAULT_COM_PORT "/dev/ttyd0"
	#endif
	#ifdef Darwin
		#define DEFAULT_COM_PORT "/dev/cu.usbserial-A30015RP"
	#endif
#endif

#include "../include/common.h"


enum Format {
	FORMAT_SREC,
	FORMAT_INHX8M,
};

struct H8DeviceInfo {
	const unsigned char *bootdata;
	size_t bootsize;
	std::vector<unsigned char> flashdata;
	std::vector<unsigned char> flashmap;
};

struct DeviceInfo {
	unsigned long id;
	char const *name;
	size_t progmemsize;
	size_t datamemsize;
	H8DeviceInfo h8;
};

struct Option {
	enum Format format;
	std::string port;
	int target_clock;
	int speed;
	std::string input_file;
	std::string output_file;
	//bool input_console;
	//bool output_console;
	DeviceInfo device;
	bool prompt;
#ifndef WIN32
	struct termios saveattr;
#endif
};

#ifdef WIN32
void sleep_ms(unsigned int ms)
{
	Sleep(ms);
}
#else
int stricmp(char const *pa, char const *pb)
{
	while (1) {
		int a = toupper((unsigned char)*pa);
		int b = toupper((unsigned char)*pb);
		if (a < b) return -1;
		if (a > b) return 1;
		if (a == 0 || b == 0) return 0;
		pa++;
		pb++;
	}
}
int strnicmp(char const *pa, char const *pb, int n)
{
	while (n > 0) {
		int a = toupper((unsigned char)*pa);
		int b = toupper((unsigned char)*pb);
		if (a < b) return -1;
		if (a > b) return 1;
		if (a == 0 || b == 0) return 0;
		pa++;
		pb++;
		n--;
	}
	return 0;
}
void sleep_ms(unsigned int ms)
{
	struct timeval to;
	to.tv_sec = 0;
	to.tv_usec = ms * 1000;
	select(0, 0, 0, 0, &to);
}
#endif



class Exception {
public:
	std::string message;
	Exception(std::string const &msg)
		: message(msg)
	{
	}
};


#ifdef WIN32
bool serial_open(Option *option, serial_handle_t *handle_p)
{
	serial_handle_t fd;
	DCB dcb;
	fd = CreateFile(option->port.c_str(), GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if (fd == INVALID_HANDLE_VALUE) {
		return false;
	}

	GetCommState(fd, &dcb);
	dcb.BaudRate = option->speed;
	dcb.fBinary = TRUE;
	dcb.fParity = FALSE;
	dcb.fOutxCtsFlow = FALSE;
	dcb.fOutxDsrFlow = FALSE;
	dcb.fDtrControl = DTR_CONTROL_ENABLE;
	dcb.fDsrSensitivity = FALSE;
	dcb.fTXContinueOnXoff = FALSE;
	dcb.fOutX = FALSE;
	dcb.fInX = FALSE;
	dcb.fErrorChar = FALSE;
	dcb.fNull = FALSE;
	dcb.fRtsControl = RTS_CONTROL_ENABLE;
	dcb.fAbortOnError = FALSE;
	dcb.ByteSize = 8;
	dcb.Parity = NOPARITY;
	dcb.StopBits = ONESTOPBIT;
	SetCommState(fd, &dcb);

	*handle_p = fd;
	return true;
}
#else
bool serial_open(Option *option, serial_handle_t *handle_p)
{
	int speed;
	serial_handle_t fd;
	struct termios attr;

	fd = open(option->port.c_str(), O_RDWR | O_NOCTTY | O_NONBLOCK);
	if (fd < 0) {
		return false;
	}

	tcgetattr(fd, &attr);
	option->saveattr = attr;

	switch (option->speed) {
//	case      50:	speed =      B50;		break;
//	case      75:	speed =      B75;		break;
//	case     110:	speed =     B110;		break;
//	case     134:	speed =     B134;		break;
//	case     150:	speed =     B150;		break;
// 	case     200:	speed =     B200;		break;
	case     300:	speed =     B300;		break;
	case     600:	speed =     B600;		break;
	case    1200:	speed =    B1200;		break;
//	case    1800:	speed =    B1800;		break;
	case    2400:	speed =    B2400;		break;
	case    4800:	speed =    B4800;		break;
	case    9600:	speed =    B9600;		break;
	case   19200:	speed =   B19200;		break;
	case   38400:	speed =   B38400;		break;
	case   57600:	speed =   B57600;		break;
	case  115200:	speed =  B115200;		break;
	case  230400:	speed =  B230400;		break;
//	case  460800:	speed =  B460800;		break;
//	case  500000:	speed =  B500000;		break;
//	case  576000:	speed =  B576000;		break;
//	case  921600:	speed =  B921600;		break;
//	case 1000000:	speed = B1000000;		break;
//	case 1152000:	speed = B1152000;		break;
//	case 1500000:	speed = B1500000;		break;
//	case 2000000:	speed = B2000000;		break;
//	case 2500000:	speed = B2500000;		break;
//	case 3000000:	speed = B3000000;		break;
//	case 3500000:	speed = B3500000;		break;
//	case 4000000:	speed = B4000000;		break;
	default:
		speed = B38400;
		break;
	}

	cfsetispeed(&attr, speed);
	cfsetospeed(&attr, speed);
	cfmakeraw(&attr);
	
	attr.c_cflag |= CS8 | CLOCAL | CREAD;
	attr.c_iflag = 0;
	attr.c_oflag = 0;
	attr.c_lflag = 0;
	attr.c_cc[VMIN] = 1;
	attr.c_cc[VTIME] = 0;

	tcsetattr(fd, TCSANOW, &attr);

	*handle_p = fd;
	return true;
}
#endif

#ifdef WIN32
void serial_close(serial_handle_t fd, Option *option)
{
	CloseHandle(fd);
}
#else
void serial_close(serial_handle_t fd, Option *option)
{
	tcsetattr(fd, TCSANOW, &option->saveattr);
	close(fd);
}
#endif

unsigned int serial_write(serial_handle_t fd, const void *ptr, unsigned long len)
{
#ifdef WIN32
	DWORD bytes = 0;
	if (WriteFile(fd, ptr, len, &bytes, NULL)) {
		return bytes;
	}
	DWORD e = GetLastError();
	return 0;
#else
	return write(fd, ptr, len);
#endif
}

unsigned int serial_read_(serial_handle_t fd, unsigned char *ptr, unsigned int len, unsigned int timeout_ms)
{
	unsigned long bytes;
#ifdef WIN32
	COMMTIMEOUTS to;
	GetCommTimeouts(fd, &to);
	to.ReadIntervalTimeout = 0;
	to.ReadTotalTimeoutMultiplier = 0;
	to.ReadTotalTimeoutConstant = timeout_ms;
	if (to.ReadTotalTimeoutConstant == 0) {
		to.ReadTotalTimeoutConstant = 1;
	}
	to.WriteTotalTimeoutMultiplier = 0;
	to.WriteTotalTimeoutConstant = 0;
	SetCommTimeouts(fd, &to);
	bytes = 0;
	if (ReadFile(fd, ptr, len, &bytes, NULL)) {
		return bytes;
	}
	return 0;
#else
	fd_set fds;
	struct timeval to;
	unsigned long n;
	unsigned char c;

	bytes = 0;
	while (len > 0) {
		FD_ZERO(&fds);
		FD_SET(fd, &fds);
		to.tv_sec = timeout_ms / 1000;
		to.tv_usec = (timeout_ms % 1000) * 1000;
		n = select(fd + 1, &fds, NULL, NULL, &to);
		if (!FD_ISSET(fd, &fds)) break;
		n = read(fd, ptr, len);
		bytes += n;
		len -= n;
		ptr += n;
	}
	return bytes;
#endif
}

int serial_getchar_(serial_handle_t fd, unsigned int timeout_ms)
{
	unsigned char c;
	if (serial_read_(fd, &c, 1, timeout_ms) == 1) {
		return c;
	}
	return -1;
}

void serial_flush_input_(serial_handle_t fd, unsigned int timeout_ms)
{
	unsigned char tmp[256];
	while (serial_read_(fd, tmp, sizeof(tmp), timeout_ms) > 0);
}

#ifndef WIN32
void noncanonical(serial_handle_t fd)
{
	struct termios tios;
	tcgetattr(fd, &tios);
	tios.c_lflag &= ~ICANON & ~ECHO;
	tios.c_cc[VTIME] = 0;
	tios.c_cc[VMIN] = 1;
	tcsetattr(fd, TCSANOW, &tios);
}
#endif









static const unsigned short crc16table[256] = {
	0x0000, 0x1189, 0x2312, 0x329b, 0x4624, 0x57ad, 0x6536, 0x74bf,
	0x8c48, 0x9dc1, 0xaf5a, 0xbed3, 0xca6c, 0xdbe5, 0xe97e, 0xf8f7,
	0x1081, 0x0108, 0x3393, 0x221a, 0x56a5, 0x472c, 0x75b7, 0x643e,
	0x9cc9, 0x8d40, 0xbfdb, 0xae52, 0xdaed, 0xcb64, 0xf9ff, 0xe876,
	0x2102, 0x308b, 0x0210, 0x1399, 0x6726, 0x76af, 0x4434, 0x55bd,
	0xad4a, 0xbcc3, 0x8e58, 0x9fd1, 0xeb6e, 0xfae7, 0xc87c, 0xd9f5,
	0x3183, 0x200a, 0x1291, 0x0318, 0x77a7, 0x662e, 0x54b5, 0x453c,
	0xbdcb, 0xac42, 0x9ed9, 0x8f50, 0xfbef, 0xea66, 0xd8fd, 0xc974,
	0x4204, 0x538d, 0x6116, 0x709f, 0x0420, 0x15a9, 0x2732, 0x36bb,
	0xce4c, 0xdfc5, 0xed5e, 0xfcd7, 0x8868, 0x99e1, 0xab7a, 0xbaf3,
	0x5285, 0x430c, 0x7197, 0x601e, 0x14a1, 0x0528, 0x37b3, 0x263a,
	0xdecd, 0xcf44, 0xfddf, 0xec56, 0x98e9, 0x8960, 0xbbfb, 0xaa72,
	0x6306, 0x728f, 0x4014, 0x519d, 0x2522, 0x34ab, 0x0630, 0x17b9,
	0xef4e, 0xfec7, 0xcc5c, 0xddd5, 0xa96a, 0xb8e3, 0x8a78, 0x9bf1,
	0x7387, 0x620e, 0x5095, 0x411c, 0x35a3, 0x242a, 0x16b1, 0x0738,
	0xffcf, 0xee46, 0xdcdd, 0xcd54, 0xb9eb, 0xa862, 0x9af9, 0x8b70,
	0x8408, 0x9581, 0xa71a, 0xb693, 0xc22c, 0xd3a5, 0xe13e, 0xf0b7,
	0x0840, 0x19c9, 0x2b52, 0x3adb, 0x4e64, 0x5fed, 0x6d76, 0x7cff,
	0x9489, 0x8500, 0xb79b, 0xa612, 0xd2ad, 0xc324, 0xf1bf, 0xe036,
	0x18c1, 0x0948, 0x3bd3, 0x2a5a, 0x5ee5, 0x4f6c, 0x7df7, 0x6c7e,
	0xa50a, 0xb483, 0x8618, 0x9791, 0xe32e, 0xf2a7, 0xc03c, 0xd1b5,
	0x2942, 0x38cb, 0x0a50, 0x1bd9, 0x6f66, 0x7eef, 0x4c74, 0x5dfd,
	0xb58b, 0xa402, 0x9699, 0x8710, 0xf3af, 0xe226, 0xd0bd, 0xc134,
	0x39c3, 0x284a, 0x1ad1, 0x0b58, 0x7fe7, 0x6e6e, 0x5cf5, 0x4d7c,
	0xc60c, 0xd785, 0xe51e, 0xf497, 0x8028, 0x91a1, 0xa33a, 0xb2b3,
	0x4a44, 0x5bcd, 0x6956, 0x78df, 0x0c60, 0x1de9, 0x2f72, 0x3efb,
	0xd68d, 0xc704, 0xf59f, 0xe416, 0x90a9, 0x8120, 0xb3bb, 0xa232,
	0x5ac5, 0x4b4c, 0x79d7, 0x685e, 0x1ce1, 0x0d68, 0x3ff3, 0x2e7a,
	0xe70e, 0xf687, 0xc41c, 0xd595, 0xa12a, 0xb0a3, 0x8238, 0x93b1,
	0x6b46, 0x7acf, 0x4854, 0x59dd, 0x2d62, 0x3ceb, 0x0e70, 0x1ff9,
	0xf78f, 0xe606, 0xd49d, 0xc514, 0xb1ab, 0xa022, 0x92b9, 0x8330,
	0x7bc7, 0x6a4e, 0x58d5, 0x495c, 0x3de3, 0x2c6a, 0x1ef1, 0x0f78
};

unsigned short crc16(unsigned short crc, unsigned char const *ptr, size_t len)
{
	for (size_t i = 0; i < len; i++) {
		unsigned char c = ptr[i];
		crc = (crc >> 8) ^ crc16table[(crc ^ c) & 0x00ff];
	}
	return crc;
}


struct write_error_info_t {
	int address;
	unsigned short write;
	unsigned short read;
};




class Connection {
private:
	serial_handle_t _serial;
	Option *_option;
public:
	Connection()
	{
		_serial = 0;
		_option = 0;
	}
public:
	bool Open(Option *option)
	{
		_option = option;
		return serial_open(_option, &_serial);
	}
	void Close()
	{
		serial_close(_serial, _option);
	}

	bool Write(unsigned char const *ptr, size_t len)
	{
		return serial_write(_serial, ptr, (unsigned long)len) == len;
	}

	bool Read(unsigned char *ptr, size_t len, unsigned int timeout_ms)
	{
		return serial_read_(_serial, ptr, (unsigned long)len, timeout_ms) == len;
	}

	void FirmwareVersion();
	void SetChipMode(unsigned char cm);
	void SetHighVoltage(bool f);

	void picCommand(int data);
	void picLoad(unsigned short data);
	unsigned short picRead();
	void picStart();
	void picLoadProgram(unsigned short c, bool byteswap = false);
	unsigned short picReadProgram(bool byteswap = false);
	void picLoadData(unsigned short c);
	unsigned short picReadData();
	void picIncrement();
	void picBeginProgram();
	void picEndProgram();

	void picBulkEraseProgramMemory(bool prompt);
	void picBulkEraseDataMemory();
	void picLoadConfig(int addr, unsigned short data);
	void picReadConfig(unsigned short *data, int len, bool endprogram = true);
	unsigned short picReadConfig(unsigned short addr);
	bool picLoadProgramWord(unsigned short data, bool byteswap = false);
	unsigned short picReadProgramWord(bool byteswap = false);
	bool picBulkLoadProgram(unsigned short const *data, int size, write_error_info_t *errorinfo);
	int picLoadProgram(const unsigned short *data, unsigned long addr, unsigned long len, bool byteswap = false, int (progress)(int addr, void *cookie) = NULL, void *cookie = NULL);
	int picReadProgram(std::vector<unsigned short> *data, int addr, int len, bool byteswap = false, int (progress)(int addr, void *cookie) = NULL, void *cookie = NULL);
	void picDisableCodeProtect();

	void avrDisable();
	void avrEnable();
	void avrLoadProgram(unsigned short addr, unsigned char size, unsigned short const *in);
	bool avrReadProgram(unsigned short addr, unsigned char size, unsigned short *out);
	unsigned char avrReadSignature(unsigned char addr);
	void avrChipErase();
	void avrReadFuse(unsigned char *ptr);
	void avrWriteFuseL(unsigned char c);
	void avrWriteFuseH(unsigned char c);
	void avrWriteFuseE(unsigned char c);

	void do_terminal();
};

#ifdef WIN32
void wait_result(HANDLE fd_ser, std::vector<unsigned char> *buffer)
{
	unsigned char tmp[1024];

	buffer->clear();

	COMMTIMEOUTS to;
	GetCommTimeouts(fd_ser, &to);
	to.ReadIntervalTimeout = 0;
	to.ReadTotalTimeoutMultiplier = 0;
	to.ReadTotalTimeoutConstant = 1000;
	to.WriteTotalTimeoutMultiplier = 0;
	to.WriteTotalTimeoutConstant = 0;
	SetCommTimeouts(fd_ser, &to);

	int len = 1;
	while (1) {
		DWORD n = 0;
		if (!ReadFile(fd_ser, tmp, len, &n, 0)) {
			return;
		}
		if (n == 0) {
			return;
		}
		buffer->insert(buffer->end(), tmp, tmp + n);
		if (buffer->at(0) < buffer->size()) {
			return;
		}
		len = 1 + buffer->at(0) - buffer->size();
	}
}
#else
void wait_result(int fd_ser, std::vector<unsigned char> *buffer)
{
	int i, n;
	unsigned char tmp[1024];
	fd_set fds;

	buffer->clear();
	while (1) {
		FD_ZERO(&fds);
		FD_SET(fd_ser, &fds);
		n = select(fd_ser + 1, &fds, NULL, NULL, NULL);
		if (!FD_ISSET(fd_ser, &fds)) {
			return;
		}
		n = read(fd_ser, tmp, sizeof(tmp));
		buffer->insert(buffer->end(), tmp, tmp + n);
		n = buffer->size();
		if (n > 0 && n == buffer->at(0) + 1) {
			return;
		}
	}
}
#endif

void calc_crc(unsigned char *ptr, size_t len)
{
	if (len < 2) {
		return;
	}
	unsigned short crc = crc16(0xffff, ptr, len - 2);
	ptr[len - 2] = (unsigned char)(crc >> 8);
	ptr[len - 1] = (unsigned char)crc;
}


void Connection::FirmwareVersion()
{
	unsigned char tmp[4];
	tmp[0] = sizeof(tmp);
	tmp[1] = _FIRMWARE_VERSION;
	calc_crc(tmp, tmp[0]);
	serial_write(_serial, tmp, tmp[0]);

	std::vector<unsigned char> buffer;
	wait_result(_serial, &buffer);

	size_t i, n;
	n = buffer.size();
	if (n > 0 && n == buffer[0] + 1) {
		for (i = 1; i < n; i++) {
			putchar(buffer[i]);
		}
	}
}

void Connection::SetChipMode(unsigned char cm)
{
	unsigned char tmp[5];
	tmp[0] = sizeof(tmp);
	tmp[1] = _SET_CHIP_MODE;
	tmp[2] = cm;
	calc_crc(tmp, tmp[0]);
	serial_write(_serial, tmp, tmp[0]);

	std::vector<unsigned char> buffer;
	wait_result(_serial, &buffer);
}

void Connection::SetHighVoltage(bool f)
{
	unsigned char tmp[5];
	tmp[0] = sizeof(tmp);
	tmp[1] = _SET_HIGH_VOLTAGE;
	tmp[2] = f;
	calc_crc(tmp, tmp[0]);
	serial_write(_serial, tmp, tmp[0]);

	std::vector<unsigned char> buffer;
	wait_result(_serial, &buffer);
}

void Connection::picCommand(int data)
{
	unsigned char tmp[5];
	tmp[0] = sizeof(tmp);
	tmp[1] = _PIC_COMMAND;
	tmp[2] = data;
	calc_crc(tmp, tmp[0]);
	serial_write(_serial, tmp, tmp[0]);

	std::vector<unsigned char> buffer;
	wait_result(_serial, &buffer);
}

void Connection::picLoad(unsigned short data)
{
	unsigned char tmp[6];
	tmp[0] = sizeof(tmp);
	tmp[1] = _PIC_LOADDATA;
	tmp[2] = (unsigned char)(data >> 8);
	tmp[3] = (unsigned char)data;
	calc_crc(tmp, tmp[0]);
	serial_write(_serial, tmp, tmp[0]);

	std::vector<unsigned char> buffer;
	wait_result(_serial, &buffer);
}

unsigned short Connection::picRead()
{
	unsigned char tmp[4];
	tmp[0] = sizeof(tmp);
	tmp[1] = _PIC_READDATA;
	calc_crc(tmp, tmp[0]);
	serial_write(_serial, tmp, tmp[0]);

	std::vector<unsigned char> buffer;
	wait_result(_serial, &buffer);

	if (buffer.size() == 3 && buffer[0] == 2) {
		return (buffer[1] << 8) | buffer[2];
	}
	return 0;
}

void Connection::picStart()
{
	unsigned char tmp[4];
	tmp[0] = sizeof(tmp);
	tmp[1] = _PIC_START;
	calc_crc(tmp, tmp[0]);
	serial_write(_serial, tmp, tmp[0]);
	std::vector<unsigned char> buffer;
	wait_result(_serial, &buffer);
}

void Connection::picEndProgram()
{
	unsigned char tmp[4];
	tmp[0] = sizeof(tmp);
	tmp[1] = _PIC_END;
	calc_crc(tmp, tmp[0]);
	serial_write(_serial, tmp, tmp[0]);

	std::vector<unsigned char> buffer;
	wait_result(_serial, &buffer);
}

void Connection::picLoadProgram(unsigned short c, bool byteswap)
{
	if (byteswap) c = (c >> 8) | (c << 8);
	picCommand(0x02);
	picLoad(c);
}

unsigned short Connection::picReadProgram(bool byteswap)
{
	picCommand(0x04);
	unsigned short c = picRead();
	if (byteswap) c = (c >> 8) | (c << 8);
	return c;
}

void Connection::picLoadData(unsigned short c)
{
	picCommand(0x03);
	picLoad(c);
}

unsigned short Connection::picReadData()
{
	picCommand(0x05);
	return picRead();
}

void Connection::picIncrement()
{
	picCommand(0x06);
}

void Connection::picBeginProgram()
{
	picCommand(0x08);
	sleep_ms(20);
}

void Connection::picBulkEraseProgramMemory(bool prompt)
{
	picStart();

	if (prompt) {
		SetHighVoltage(true);
		puts("Press enter.");
		getchar();
	}

	picCommand(0x00); // sets the PC to 0x2000
	picLoadProgram(0x3fff, false);
	picCommand(0x09);
	sleep_ms(10);
	picEndProgram();
}

void Connection::picBulkEraseDataMemory()
{
	picStart();
	picCommand(0x0b);
	sleep_ms(10);
	picEndProgram();
}

void Connection::picLoadConfig(int addr, unsigned short data)
{
	picStart();

	// dummy read for 16F628
	{
		picCommand(0x00);
		for (int i = 0; i < 4; i++) {
			if (i > 0) {
				picIncrement();
			}
			picCommand(0x04); // read program
			{
				unsigned char tmp[4];
				tmp[0] = sizeof(tmp);
				tmp[1] = _PIC_READDATA;
				calc_crc(tmp, tmp[0]);
				serial_write(_serial, tmp, tmp[0]);
				std::vector<unsigned char> buffer;
				wait_result(_serial, &buffer);
			}
		}
	}

	picCommand(0x00);
	picLoad(data);
	for (int a = 0; a < addr; a++) {
		picIncrement();
	}
	picBeginProgram();
	picEndProgram();
}

void Connection::picReadConfig(unsigned short *data, int len, bool endprogram)
{
	picStart();

	picCommand(0x00);
	for (int i = 0; i < len; i++) {
		if (i > 0) {
			picIncrement();
		}
		picCommand(0x04); // read program
		unsigned short w = 0x3fff;
		{
			unsigned char tmp[4];
			tmp[0] = sizeof(tmp);
			tmp[1] = _PIC_READDATA;
			calc_crc(tmp, tmp[0]);
			serial_write(_serial, tmp, tmp[0]);
			std::vector<unsigned char> buffer;
			wait_result(_serial, &buffer);
			if (buffer.size() == 3 && buffer[0] == 2) {
				w = (buffer[1] << 8) | buffer[2];
			}
		}
		data[i] = w;
	}
	if (endprogram) {
		picEndProgram();
	}
}

unsigned short Connection::picReadConfig(unsigned short addr)
{
	picStart();
	picCommand(0x00);
	while (addr > 0) {
		picReadProgram();
		picIncrement();
		addr--;
	}
	unsigned short data = picReadProgram();
	picEndProgram();
	return data;
}

bool Connection::picLoadProgramWord(unsigned short data, bool byteswap)
{
	unsigned short c;

	if (byteswap) data = (data >> 8) | (data << 8);

	picLoadProgram(data);
	picBeginProgram();
	c = picReadProgram();

	if ((c ^ data) & 0x3fff) {
		return false;
	}
	return true;
}
unsigned short Connection::picReadProgramWord(bool byteswap)
{
	unsigned short c = picReadProgram();
	if (byteswap) c = (c >> 8) | (c << 8);
	return c;
}

bool Connection::picBulkLoadProgram(unsigned short const *data, int size, write_error_info_t *errorinfo)
{
	int i;
	unsigned char tmp[6 + 2 * 120 + 2];
	if (size > 120) {
		size = 120;
	}
	tmp[0] = 6 + 2 * size + 2;
	tmp[1] = _PIC_BULK_LOAD_PROGRAM;
	tmp[2] = 0;		// reserved
	tmp[3] = 0;		// reserved
	tmp[4] = size;
	tmp[5] = 0;		// reserved
	for (i = 0; i < size; i++) {
		tmp[i * 2 + 6] = (unsigned char)(data[i] >> 8);
		tmp[i * 2 + 7] = (unsigned char)data[i];
	}
	calc_crc(tmp, tmp[0]);
	serial_write(_serial, tmp, tmp[0]);

	std::vector<unsigned char> buffer;
	wait_result(_serial, &buffer);
	if (buffer.size() == 1 && buffer[0] == 0) {
		return true;
	}
	if (buffer.size() == 6 && buffer[0] == 5) {
		if (errorinfo) {
			errorinfo->address = buffer[1];
			errorinfo->write = (buffer[2] << 8) | buffer[3];
			errorinfo->read = (buffer[4] << 8) | buffer[5];
		}
	}
	return false;
}


int Connection::picLoadProgram(const unsigned short *data, unsigned long addr, unsigned long len, bool byteswap, int (*progress)(int addr, void *cookie), void *cookie)
{
	size_t i;
	int re;

	re = 0;
	picStart();

	for (i = 0; i < addr; i++)
		if (i) picIncrement();

	for (i = 0; i < len; i++) {
		if (i) picIncrement();

		if (!picLoadProgramWord(*data, byteswap)) {
			picEndProgram();
			return -1;
		}

		if (progress) {
			re = progress((int)i, cookie);
			if (re) break;
		}
		addr++;
		data++;
	}
	picEndProgram();
	return re;
}

int Connection::picReadProgram(std::vector<unsigned short> *data, int addr, int len, bool byteswap, int (*progress)(int addr, void *cookie), void *cookie)
{
	int i, re;

	re = 0;
	picStart();
	
	for (i = 0; i < addr; i++)
		picIncrement();

	for (i = 0; i < len; i++) {
		if (i) picIncrement();

		data->push_back(picReadProgramWord(byteswap));

		if (progress) {
			re = progress(i, cookie);
			if (re) break;
		}
		addr++;
	}
	picEndProgram();
	return re;
}

void Connection::picDisableCodeProtect()
{
	int i;
	picStart();
	picCommand(0x00);
	picLoad(0x3fff);
	for (i = 0; i < 7; i++) {
		picIncrement();
	}
	picCommand(0x01);
	picCommand(0x07);
	picBeginProgram();
	picCommand(0x01);
	picCommand(0x07);
	picEndProgram();
}


void Connection::avrDisable()
{
	unsigned char tmp[4];
	tmp[0] = sizeof(tmp);
	tmp[1] = _AVR_DISABLE;
	calc_crc(tmp, tmp[0]);
	serial_write(_serial, tmp, tmp[0]);

	std::vector<unsigned char> buffer;
	wait_result(_serial, &buffer);
}

void Connection::avrEnable()
{
	unsigned char tmp[4];
	tmp[0] = sizeof(tmp);
	tmp[1] = _AVR_ENABLE;
	calc_crc(tmp, tmp[0]);
	serial_write(_serial, tmp, tmp[0]);

	std::vector<unsigned char> buffer;
	wait_result(_serial, &buffer);
}

void Connection::avrLoadProgram(unsigned short addr, unsigned char size, unsigned short const *in)
{
	int i;
	unsigned char tmp[6 + 2 * 120 + 2];
	if (size > 120) {
		size = 120;
	}
	tmp[0] = 6 + 2 * size + 2;
	tmp[1] = _AVR_LOAD_PROGRAM;
	tmp[2] = (unsigned char)(addr >> 8);
	tmp[3] = (unsigned char)addr;
	tmp[4] = size;
	tmp[5] = 0;		// reserved
	for (i = 0; i < size; i++) {
		tmp[i * 2 + 6] = (unsigned char)(in[i] >> 8);
		tmp[i * 2 + 7] = (unsigned char)in[i];
	}
	calc_crc(tmp, tmp[0]);
	serial_write(_serial, tmp, tmp[0]);

	std::vector<unsigned char> buffer;
	wait_result(_serial, &buffer);
}

bool Connection::avrReadProgram(unsigned short addr, unsigned char size, unsigned short *out)
{
	unsigned char tmp[7];
	if (size > 16) {
		size = 16;
	}
	tmp[0] = sizeof(tmp);
	tmp[1] = _AVR_READ_PROGRAM;
	tmp[2] = (unsigned char)(addr >> 8);
	tmp[3] = (unsigned char)addr;
	tmp[4] = size;
	calc_crc(tmp, tmp[0]);
	serial_write(_serial, tmp, tmp[0]);

	std::vector<unsigned char> buffer;
	wait_result(_serial, &buffer);

	if (buffer.size() == size * 2 + 1 && buffer[0] == size * 2) {
		int i;
		for (i = 0; i < size; i++) {
			out[i] = (buffer[i * 2 + 1] << 8) | buffer[i * 2 + 2];
		}
		return true;
	}
	return false;
}

unsigned char Connection::avrReadSignature(unsigned char addr)
{
	unsigned char tmp[5];
	tmp[0] = sizeof(tmp);
	tmp[1] = _AVR_READ_SIGNATURE;
	tmp[2] = addr;
	calc_crc(tmp, tmp[0]);
	serial_write(_serial, tmp, tmp[0]);

	std::vector<unsigned char> buffer;
	wait_result(_serial, &buffer);

	if (buffer.size() == 2 && buffer[0] == 1) {
		return buffer[1];
	}
	return 0;
}

void Connection::avrChipErase()
{
	unsigned char tmp[4];
	tmp[0] = sizeof(tmp);
	tmp[1] = _AVR_CHIP_ERASE;
	calc_crc(tmp, tmp[0]);
	serial_write(_serial, tmp, tmp[0]);

	std::vector<unsigned char> buffer;
	wait_result(_serial, &buffer);
}

void Connection::avrReadFuse(unsigned char *ptr)
{
	unsigned char tmp[4];
	tmp[0] = sizeof(tmp);
	tmp[1] = _AVR_READ_FUSE;
	calc_crc(tmp, tmp[0]);
	serial_write(_serial, tmp, tmp[0]);

	std::vector<unsigned char> buffer;
	wait_result(_serial, &buffer);

	if (buffer.size() == 5 && buffer[0] == 4) {
		memcpy(ptr, &buffer[1], 4);
	} else {
		fprintf(stderr, "avrReadFuse: falied\n");
	}
}

void Connection::avrWriteFuseL(unsigned char c)
{
	unsigned char tmp[5];
	tmp[0] = sizeof(tmp);
	tmp[1] = _AVR_WRITE_FUSE_L;
	tmp[2] = c;
	calc_crc(tmp, tmp[0]);
	serial_write(_serial, tmp, tmp[0]);

	std::vector<unsigned char> buffer;
	wait_result(_serial, &buffer);
}

void Connection::avrWriteFuseH(unsigned char c)
{
	unsigned char tmp[5];
	tmp[0] = sizeof(tmp);
	tmp[1] = _AVR_WRITE_FUSE_H;
	tmp[2] = c;
	calc_crc(tmp, tmp[0]);
	serial_write(_serial, tmp, tmp[0]);

	std::vector<unsigned char> buffer;
	wait_result(_serial, &buffer);
}

void Connection::avrWriteFuseE(unsigned char c)
{
	unsigned char tmp[5];
	tmp[0] = sizeof(tmp);
	tmp[1] = _AVR_WRITE_FUSE_E;
	tmp[2] = c;
	calc_crc(tmp, tmp[0]);
	serial_write(_serial, tmp, tmp[0]);

	std::vector<unsigned char> buffer;
	wait_result(_serial, &buffer);
}



#ifdef WIN32


class Win32Terminal {
private:
	HANDLE _h_event;
	HANDLE _h_mutex;
	serial_handle_t _h_com;
	HANDLE _h_con_thread;
	HANDLE _h_com_thread;
	HANDLE _h_conin;
	std::vector<unsigned char> _con_in_buffer;
	std::vector<unsigned char> _com_in_buffer;
	void ConInputThread()
	{
		unsigned char tmp[1024];
		DWORD n = 0;
		while (1) {
			if (!ReadFile(_h_conin, tmp, sizeof(tmp), &n, 0)) {
				break;
			}
			WaitForSingleObject(_h_mutex, INFINITE);
			_con_in_buffer.insert(_con_in_buffer.end(), tmp, tmp + n);
			SetEvent(_h_event);
			ReleaseMutex(_h_mutex);
		}
	}
	static unsigned int __stdcall con_input_thread(void *cookie)
	{
		Win32Terminal *me = (Win32Terminal *)cookie;
		me->ConInputThread();
		return 0;
	}
	void ComInputThread()
	{
		unsigned char tmp[1024];
		DWORD n = 0;
		while (1) {
			int n = serial_read_(_h_com, tmp, sizeof(tmp), 1);
			WaitForSingleObject(_h_mutex, INFINITE);
			_com_in_buffer.insert(_com_in_buffer.end(), tmp, tmp + n);
			SetEvent(_h_event);
			ReleaseMutex(_h_mutex);
		}
	}
	static unsigned int __stdcall com_input_thread(void *cookie)
	{
		Win32Terminal *me = (Win32Terminal *)cookie;
		me->ComInputThread();
		return 0;
	}
public:
	void start(Option *option)
	{
		_h_event = CreateEvent(0, FALSE, FALSE, 0);
		_h_mutex = CreateMutex(0, FALSE, 0);
		if (!serial_open(option, &_h_com)) {
			fprintf(stderr, "Could not open device '%s'\n", option->port.c_str());
			return;
		}
		_h_conin = CreateFile("CONIN$", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, 0, OPEN_EXISTING, 0, 0);
		SetConsoleMode(_h_conin, 0);
		_h_con_thread = (HANDLE)_beginthreadex(0, 0, con_input_thread, this, CREATE_SUSPENDED, 0);
		_h_com_thread = (HANDLE)_beginthreadex(0, 0, com_input_thread, this, CREATE_SUSPENDED, 0);
		ResumeThread(_h_con_thread);
		ResumeThread(_h_com_thread);

		while (1) {
			WaitForSingleObject(_h_event, INFINITE);
			WaitForSingleObject(_h_mutex, INFINITE);
			if (!_com_in_buffer.empty()) {
				fwrite(&_com_in_buffer[0], 1, _com_in_buffer.size(), stdout);
				_com_in_buffer.clear();
			}
			if (!_con_in_buffer.empty()) {
				serial_write(_h_com, &_con_in_buffer[0], (unsigned long)_con_in_buffer.size());
				_con_in_buffer.clear();
			}
			ReleaseMutex(_h_mutex);
		}

	}
};

#include <signal.h>

void Connection::do_terminal()
{
	Win32Terminal term;


	term.start(_option);
}

#else
void Connection::do_terminal()
{
	int n, fd_tty, fd_ser;
	fd_set fds;
	struct termios tios_tty;

	if (!serial_open(_option, &fd_ser)) {
		return;
	}

	fd_tty = 0;		// stdin
	tcgetattr(fd_tty, &tios_tty);
	noncanonical(fd_tty);
#if 1
	unsigned char tmp[4];
	tmp[0] = sizeof(tmp);
	tmp[1] = _FIRMWARE_VERSION;
	calc_crc(tmp, tmp[0]);
	serial_write(_serial, tmp, tmp[0]);
#endif
	while (1) {
		FD_ZERO(&fds);
		FD_SET(fd_tty, &fds);		// stdin
		FD_SET(fd_ser, &fds);		// serial
		n = select(fd_ser + 1, &fds, NULL, NULL, NULL);

		if (FD_ISSET(fd_tty, &fds)) {		// transmit
			int c = getchar();
			if (c == '\n') c = '\r';
			if (c >= 0) {
				unsigned char t = c;
				serial_write(fd_ser, &t, 1);
			}
		}

		if (FD_ISSET(fd_ser, &fds)) {	// receive
			unsigned char c;
			if (read(fd_ser, &c, 1) == 1) {
				putchar(c);
				fflush(stdout);
			}
		}
	}

	serial_close(fd_ser, _option);
}
#endif



////////////////////////////////////////////////////////////////////////////////
struct bin_t {
	unsigned short addr;
	std::vector<unsigned char> data;
};

class XP {
private:
	Option _option;
	Connection _connection;

	struct OptionTable {
		bool (XP::*handler)(const char *arg);
		char *name;
	};
	static OptionTable _options[];


	bool option_read(const char *arg);
	bool option_write(const char *arg);
	bool option_verify(const char *arg);
	bool option_format(const char *arg);
	bool option_confword(const char *arg);
	bool option_test(const char *arg);
	bool option_erase(const char *arg);

	void output_binary_le(FILE *fp, const unsigned short *ptr, size_t addr, size_t size);
	void output_binary_be(FILE *fp, const unsigned short *ptr, size_t addr, size_t size);
	void output_mot(FILE *fp, const unsigned short *ptr, size_t addr, size_t size);
	void output_inhx8m(FILE *fp, const unsigned short *ptr, size_t addr, size_t size);

	bool pic_bulk_write(std::list<bin_t> const *bin, write_error_info_t *errorinfo);

	bool receive_h8_3069_response(unsigned char *dst, int timeout = 1000);
	bool write_h8_3069_128bytes(unsigned long addr, unsigned char const *src);

	void input_binary_le(FILE *fp);
	void input_binary_be(FILE *fp);
	void input_mot_write(FILE *fp);
	void input_inhx8m_write(FILE *fp);

	void FirmwareVersion();

	void Read();
	void Write();

	void pic_inquire();
	void pic_disable_code_protect();
	void pic_read_confword();
	void pic_write_confword(unsigned short data);
	void pic_read_p(unsigned short addr);
	void pic_write_p(unsigned short addr, unsigned short data);
	void pic_bulk_erase_program_memory();
	void pic_bulk_erase_data_memory();
	void pic_erase_pic12f629();

	void avr_inquire();
	void avr_program_mode(bool f);
	void avr_chip_erase();
	void avr_read_fuse();
	void avr_write_fuse_l(unsigned char c);
	void avr_write_fuse_h(unsigned char c);
	void avr_write_fuse_e(unsigned char c);
	bool avr_write(unsigned short addr, int n, unsigned short const *data, write_error_info_t *errorinfo);
	bool avr_write(unsigned short addr, int n, unsigned char const *data, write_error_info_t *errorinfo);

	void do_test();

	bool GetDeviceInfo(char const *deviceid, DeviceInfo *deviceinfo);
	bool GetDeviceInfo(unsigned long deviceid, DeviceInfo *deviceinfo);

	bool OpenPort(); // throws exception
	void ClosePort();

	bool h8_transfer_boot_program(const unsigned char *loader, size_t length);

public:
	XP();

	int main(int argc, char **argv);
};

XP::XP()
{
	_option.format = FORMAT_INHX8M;
	_option.port = DEFAULT_COM_PORT;
	_option.speed = 38400;
	_option.device.id = 0;
	_option.device.name = "";
	_option.device.progmemsize = 0;
	_option.device.datamemsize = 0;
	_option.target_clock = 25000000;
	_option.prompt = false;
	//_option.input_console = false;
	//_option.output_console = false;
}


bool XP::OpenPort()
{
	return _connection.Open(&_option);
}
void XP::ClosePort()
{
	_connection.Close();
}

#ifdef DEVICE_TYPE
#undef DEVICE_TYPE
#endif

#define PIC_ID(X) (0x00010000 | X)
#define AVR_ID(X) (0x00020000 | X)
#define H8_ID(X) (0x00030000 | X)
#define DEVICE_TYPE(X) (X & 0xffff0000)
enum {
	DEVICE_TYPE_PIC = 0x00010000,
	DEVICE_TYPE_AVR = 0x00020000,
	DEVICE_TYPE_H8  = 0x00030000,
};

#include "h8/bootloader.h"

bool XP::GetDeviceInfo(char const *name, DeviceInfo *deviceinfo)
{
	static struct {
		unsigned long id;
		char const *name;
		size_t progmemsize;
		size_t datamemsize;
		const unsigned char *bootdata;
		size_t bootsize;
	} table[] = {
		PIC_ID(0x0560), "PIC16F84",                 1024,       64,              0,            0,
		PIC_ID(0)     , "PIC16C84",                 1024,       64,              0,            0,
		PIC_ID(0)     , "PIC16F84",                 1024,       64,              0,            0,
		PIC_ID(0)     , "PIC16CR84",                1024,       64,              0,            0,
		PIC_ID(0)     , "PIC16F83",                  512,       64,              0,            0,
		PIC_ID(0)     , "PIC16CR83",                 512,       64,              0,            0,

		PIC_ID(0x0f80), "PIC12F629",                1024,      128,              0,            0,
		PIC_ID(0x0fc0), "PIC12F675",                1024,      128,              0,            0,
		PIC_ID(0x1fc0), "PIC16F630",                1024,      128,              0,            0,
		PIC_ID(0x1fe0), "PIC16F676",                1024,      128,              0,            0,

		PIC_ID(0x1040), "PIC16F627",                1024,      128,              0,            0,
		PIC_ID(0x1060), "PIC16F628",                2048,      128,              0,            0,
		PIC_ID(0x1100), "PIC16F648",                4096,      256,              0,            0,

		PIC_ID(0x2000), "PIC16F882",                   0,        0,              0,            0,
		PIC_ID(0x2020), "PIC16F883",                   0,        0,              0,            0,
		PIC_ID(0x2040), "PIC16F884",                   0,        0,              0,            0,
		PIC_ID(0x2060), "PIC16F886",                   0,        0,              0,            0,
		PIC_ID(0x2080), "PIC16F887",                   0,        0,              0,            0,

		PIC_ID(0x2180), "PIC12F615",                1024,        0,              0,            0,
		PIC_ID(0x21a0), "PIC12HV615",               1024,        0,              0,            0,
		PIC_ID(0x1240), "PIC16F616",                2048,        0,              0,            0,
		PIC_ID(0x1260), "PIC16HV616",               2048,        0,              0,            0,
		PIC_ID(0x2240), "PIC12F609",                1024,        0,              0,            0,
		PIC_ID(0x2280), "PIC12HV609",               1024,        0,              0,            0,
		PIC_ID(0x2260), "PIC16F610",                1024,        0,              0,            0,
		PIC_ID(0x22a0), "PIC16HV610",               1024,        0,              0,            0,

		PIC_ID(0x1380), "PIC16F917",                8192,      256,              0,            0,
		PIC_ID(0x13a0), "PIC16F916",                8192,      256,              0,            0,
		PIC_ID(0x13c0), "PIC16F914",                4096,      256,              0,            0,
		PIC_ID(0x13e0), "PIC16F913",                4096,      256,              0,            0,
		PIC_ID(0x1460), "PIC16F946",                8192,      256,              0,            0,

		PIC_ID(0x0fa0), "PIC12F635",                1024,      128,              0,            0,
		PIC_ID(0x0460), "PIC12F683",                2048,      256,              0,            0,
		PIC_ID(0x1420), "PIC16F631",                1024,      128,              0,            0,
		PIC_ID(0x10a0), "PIC16F636",                2048,      256,              0,            0,
		PIC_ID(0x10a0), "PIC16F639",                2048,      256,              0,            0,
		PIC_ID(0x1440), "PIC16F677",                2048,      256,              0,            0,
		PIC_ID(0x1080), "PIC16F684",                2048,      256,              0,            0,
		PIC_ID(0x04a0), "PIC16F685",                4096,      256,              0,            0,
		PIC_ID(0x1320), "PIC16F687",                2048,      256,              0,            0,
		PIC_ID(0x1180), "PIC16F688",                4096,      256,              0,            0,
		PIC_ID(0x1340), "PIC16F689",                4096,      256,              0,            0,
		PIC_ID(0x1400), "PIC16F690",                4096,      256,              0,            0,

		PIC_ID(0x1220), "PIC16HV785",               2048,      256,              0,            0,
		PIC_ID(0x1200), "PIC16F785",                2048,      256,              0,            0,

		PIC_ID(0x0ba0), "PIC16F737",                4096,        0,              0,            0,
		PIC_ID(0x0be0), "PIC16F747",                4096,        0,              0,            0,
		PIC_ID(0x0ea0), "PIC16F767",                8192,        0,              0,            0,
		PIC_ID(0x0de0), "PIC16F777",                8192,        0,              0,            0,

		PIC_ID(0x04c0), "PIC16F818",                1024,      128,              0,            0,
		PIC_ID(0x04e0), "PIC16F819",                2048,      256,              0,            0,

		PIC_ID(0x0720), "PIC16F87",                 4096,      256,              0,            0,
		PIC_ID(0x0760), "PIC16F88",                 4096,      256,              0,            0,

		PIC_ID(0x1140), "PIC16F716",                2048,        0,              0,            0,

		PIC_ID(0x0600), "PIC16F73",                 4096,        0,              0,            0,
		PIC_ID(0x0620), "PIC16F74",                 4096,        0,              0,            0,
		PIC_ID(0x0640), "PIC16F76",                 8192,        0,              0,            0,
		PIC_ID(0x0660), "PIC16F77",                 8192,        0,              0,            0,

		PIC_ID(0x0d00), "PIC16F870",                2048,       64,              0,            0,
		PIC_ID(0x0d20), "PIC16F871",                2048,       64,              0,            0,
		PIC_ID(0x08e0), "PIC16F872",                2048,       64,              0,            0,
		PIC_ID(0x0960), "PIC16F873",                4096,      128,              0,            0,
		PIC_ID(0x0920), "PIC16F874",                4096,      128,              0,            0,
		PIC_ID(0x09e0), "PIC16F876",                8192,      256,              0,            0,
		PIC_ID(0x09a0), "PIC16F877",                8192,      256,              0,            0,

		PIC_ID(0x0e40), "PIC16F873",                   0,        0,              0,            0,
		PIC_ID(0x0e60), "PIC16F874",                   0,        0,              0,            0,
		PIC_ID(0x0e00), "PIC16F876",                   0,        0,              0,            0,
		PIC_ID(0x0e20), "PIC16F877",                   0,        0,              0,            0,

		PIC_ID(0x07a0), "PIC16F627",                1024,      128,              0,            0,
		PIC_ID(0x07c0), "PIC16F628",                2048,      128,              0,            0,

		AVR_ID(0x9101), "AT90S2313",                1024,      128,              0,            0,
		AVR_ID(0x910a), "ATTINY2313",               1024,      128,              0,            0,
		AVR_ID(0x9205), "ATMEGA48",                 4096,      256,              0,            0,
		AVR_ID(0x930a), "ATMEGA88",                 8192,      512,              0,            0,
		AVR_ID(0x9406), "ATMEGA168",               16384,      512,              0,            0,
		AVR_ID(0x9205), "ATMEGA48",                 4096,      256,              0,            0,
		AVR_ID(0x930a), "ATMEGA88",                 8192,      512,              0,            0,
		AVR_ID(0x9406), "ATMEGA168",               16384,      512,              0,            0,
		AVR_ID(0x950f), "ATMEGA328",               32768,     1024,              0,            0,

		H8_ID(0x3664),  "H8/3664",                0x8000,        0, boot3664_data, boot3664_size,
		H8_ID(0x3048),  "H8/3048",                0x8000,        0, boot3048_data, boot3048_size,
		H8_ID(0x3052),  "H8/3052",               0x80000,        0, boot3052_data, boot3052_size,
		H8_ID(0x3067),  "H8/3067",                0x8000,        0, boot3067_data, boot3067_size,
		H8_ID(0x3068),  "H8/3068",               0x60000,        0, boot3068_data, boot3068_size,
		H8_ID(0x3069),  "H8/3069",               0x80000,        0, boot3068_data, boot3068_size,
	};
	int i, n;
	n = sizeof(table) / sizeof(*table);
	for (i = 0; i < n; i++) {
		if (stricmp(name, table[i].name) == 0) {
			deviceinfo->id = table[i].id;
			deviceinfo->name = table[i].name;
			deviceinfo->progmemsize = table[i].progmemsize;
			deviceinfo->datamemsize = table[i].datamemsize;
			deviceinfo->h8.bootdata = table[i].bootdata;
			deviceinfo->h8.bootsize = table[i].bootsize;
			return true;
		}
	}
	return false;
}

bool XP::GetDeviceInfo(unsigned long id, DeviceInfo *deviceinfo)
{
	char const *name = 0;

	switch (id) {
	case PIC_ID(0x0560):  name = "PIC16F84";                   break;

	case PIC_ID(0x0f80):  name = "PIC12F629";                   break;
	case PIC_ID(0x0fc0):  name = "PIC12F675";                   break;
	case PIC_ID(0x1fc0):  name = "PIC16F630";                   break;
	case PIC_ID(0x1fe0):  name = "PIC16F676";                   break;

	case PIC_ID(0x1040):  name = "PIC16F627";                  break;
	case PIC_ID(0x1060):  name = "PIC16F628";                  break;
	case PIC_ID(0x1100):  name = "PIC16F648";                  break;

	case PIC_ID(0x2000):  name = "PIC16F882";                   break;
	case PIC_ID(0x2020):  name = "PIC16F883";                   break;
	case PIC_ID(0x2040):  name = "PIC16F884";                   break;
	case PIC_ID(0x2060):  name = "PIC16F886";                   break;
	case PIC_ID(0x2080):  name = "PIC16F887";                   break;

	case PIC_ID(0x2180):  name = "PIC12F615";                   break;
	case PIC_ID(0x21a0):  name = "PIC12HV615";                  break;
	case PIC_ID(0x1240):  name = "PIC16F616";                   break;
	case PIC_ID(0x1260):  name = "PIC16HV616";                  break;
	case PIC_ID(0x2240):  name = "PIC12F609";                   break;
	case PIC_ID(0x2280):  name = "PIC12HV609";                  break;
	case PIC_ID(0x2260):  name = "PIC16F610";                   break;
	case PIC_ID(0x22a0):  name = "PIC16HV610";                  break;

	case PIC_ID(0x1380):  name = "PIC16F917";                   break;
	case PIC_ID(0x13a0):  name = "PIC16F916";                   break;
	case PIC_ID(0x13c0):  name = "PIC16F914";                   break;
	case PIC_ID(0x13e0):  name = "PIC16F913";                   break;
	case PIC_ID(0x1460):  name = "PIC16F946";                   break;

	case PIC_ID(0x0fa0):  name = "PIC12F635";                   break;
	case PIC_ID(0x0460):  name = "PIC12F683";                   break;
	case PIC_ID(0x1420):  name = "PIC16F631";                   break;
	case PIC_ID(0x10a0):  name = "PIC16F636/PIC16F639";         break;
	case PIC_ID(0x1440):  name = "PIC16F677";                   break;
	case PIC_ID(0x1080):  name = "PIC16F684";                   break;
	case PIC_ID(0x04a0):  name = "PIC16F685";                   break;
	case PIC_ID(0x1320):  name = "PIC16F687";                   break;
	case PIC_ID(0x1180):  name = "PIC16F688";                   break;
	case PIC_ID(0x1340):  name = "PIC16F689";                   break;
	case PIC_ID(0x1400):  name = "PIC16F690";                   break;

	case PIC_ID(0x1220):  name = "PIC16HV785";                  break;
	case PIC_ID(0x1200):  name = "PIC16F785";                   break;

	case PIC_ID(0x0ba0):  name = "PIC16F737";                   break;
	case PIC_ID(0x0be0):  name = "PIC16F747";                   break;
	case PIC_ID(0x0ea0):  name = "PIC16F767";                   break;
	case PIC_ID(0x0de0):  name = "PIC16F777";                   break;

	case PIC_ID(0x04c0):  name = "PIC16F818";                   break;
	case PIC_ID(0x04e0):  name = "PIC16F819";                   break;

	case PIC_ID(0x0720):  name = "PIC16F87";                    break;
	case PIC_ID(0x0760):  name = "PIC16F88";                    break;

	case PIC_ID(0x1140):  name = "PIC16F716";                   break;

	case PIC_ID(0x0600):  name = "PIC16F73";                    break;
	case PIC_ID(0x0620):  name = "PIC16F74";                    break;
	case PIC_ID(0x0640):  name = "PIC16F76";                    break;
	case PIC_ID(0x0660):  name = "PIC16F77";                    break;

	case PIC_ID(0x0e40):  name = "PIC16F873";                  break;
	case PIC_ID(0x0e60):  name = "PIC16F874";                  break;
	case PIC_ID(0x0e00):  name = "PIC16F876";                  break;
	case PIC_ID(0x0e20):  name = "PIC16F877";                  break;

	case PIC_ID(0x07a0):  name = "PIC16F627";                   break;
	case PIC_ID(0x07c0):  name = "PIC16F628";                   break;

	case PIC_ID(0x0d00):  name = "PIC16F870";                   break;
	case PIC_ID(0x0d20):  name = "PIC16F871";                   break;
	case PIC_ID(0x08e0):  name = "PIC16F872";                   break;
	case PIC_ID(0x0960):  name = "PIC16F873";                   break;
	case PIC_ID(0x0920):  name = "PIC16F874";                   break;
	case PIC_ID(0x09e0):  name = "PIC16F876";                   break;
	case PIC_ID(0x09a0):  name = "PIC16F877";                   break;

	case AVR_ID(0x9101):  name = "AT90S2313";                   break;
	case AVR_ID(0x910a):  name = "ATTINY2313";                  break;
	case AVR_ID(0x9205):  name = "ATMEGA48";                    break;
	case AVR_ID(0x930a):  name = "ATMEGA88";                    break;
	case AVR_ID(0x9406):  name = "ATMEGA168";                   break;
	case AVR_ID(0x950f):  name = "ATMEGA328";                   break;

	case H8_ID(0x3664):   name = "H8/3664";                     break;
	case H8_ID(0x3048):   name = "H8/3048";                     break;
	case H8_ID(0x3067):   name = "H8/3067";                     break;
	case H8_ID(0x3068):   name = "H8/3068";                     break;
	case H8_ID(0x3069):   name = "H8/3069";                     break;

	default:
		return false;
	}
	deviceinfo->id = id;
	deviceinfo->name = name;
	deviceinfo->progmemsize = 0;
	deviceinfo->datamemsize = 0;
	return GetDeviceInfo(name, deviceinfo);
}

void XP::pic_inquire()
{
	unsigned short array[16];

	_connection.picReadConfig(array, 16);

	for (int i = 0; i < 16; i++) {
		if (i > 0) {
			putchar(' ');
		}
		printf("  +%X", i);

	}
	putchar('\n');

	for (int i = 0; i < 16; i++) {
		if (i > 0) {
			putchar(' ');
		}
		printf("----", i);

	}
	putchar('\n');

	for (int i = 0; i < 16; i++) {
		if (i > 0) {
			putchar(' ');
		}
		printf("%04X", array[i]);
	}
	putchar('\n');

	putchar('\n');

	unsigned long deviceid = PIC_ID(array[6] & 0x3fe0);

	DeviceInfo deviceinfo;
	bool unknown = !GetDeviceInfo(deviceid, &deviceinfo);

	printf("Device: ");
	if (unknown) {
		puts("Unknown");
		return;
	}

	puts(deviceinfo.name);

	putchar('\n');

	switch (deviceid) {
	case PIC_ID(0x2000): // PIC16F882
	case PIC_ID(0x2020): // PIC16F883
	case PIC_ID(0x2040): // PIC16F884
	case PIC_ID(0x2060): // PIC16F886
	case PIC_ID(0x2080): // PIC16F887
		printf("   DEBUG: %c\n", (array[7] & 0x2000) ? '1' : '0');
		printf("     LCP: %c\n", (array[7] & 0x1000) ? '1' : '0');
		printf("   FCMEN: %c\n", (array[7] & 0x0800) ? '1' : '0');
		printf("    IESO: %c\n", (array[7] & 0x0400) ? '1' : '0');
		printf("  BOREN1: %c\n", (array[7] & 0x0200) ? '1' : '0');
		printf("  BOREN0: %c\n", (array[7] & 0x0100) ? '1' : '0');
		printf("     CPD: %c\n", (array[7] & 0x0080) ? '1' : '0');
		printf("      CP: %c\n", (array[7] & 0x0040) ? '1' : '0');
		printf("   MCLRE: %c\n", (array[7] & 0x0020) ? '1' : '0');
		printf("   PWRTE: %c\n", (array[7] & 0x0010) ? '1' : '0');
		printf("   WDTEN: %c\n", (array[7] & 0x0008) ? '1' : '0');
		printf("   FOSC2: %c\n", (array[7] & 0x0004) ? '1' : '0');
		printf("   FOSC1: %c\n", (array[7] & 0x0002) ? '1' : '0');
		printf("   FOSC0: %c\n", (array[7] & 0x0001) ? '1' : '0');
		break;
	case PIC_ID(0x1040): // PIC16F627A
	case PIC_ID(0x1060): // PIC16F628A
	case PIC_ID(0x1100): // PIC16F648A
		printf("      CP: %c\n", (array[7] & 0x2000) ? '1' : '0');
		printf("     CPD: %c\n", (array[7] & 0x0100) ? '1' : '0');
		printf("     LVP: %c\n", (array[7] & 0x0080) ? '1' : '0');
		printf("   BODEN: %c\n", (array[7] & 0x0040) ? '1' : '0');
		printf("   MCLRE: %c\n", (array[7] & 0x0020) ? '1' : '0');
		printf("   FOSC2: %c\n", (array[7] & 0x0010) ? '1' : '0');
		printf("   PWRTE: %c\n", (array[7] & 0x0008) ? '1' : '0');
		printf("    WDTE: %c\n", (array[7] & 0x0004) ? '1' : '0');
		printf("   FOSC1: %c\n", (array[7] & 0x0002) ? '1' : '0');
		printf("   FOSC0: %c\n", (array[7] & 0x0001) ? '1' : '0');
		break;
	case PIC_ID(0x2180): // PIC12F615
	case PIC_ID(0x21a0): // PIC12HV615
	case PIC_ID(0x1240): // PIC16F616
	case PIC_ID(0x1260): // PIC16HV616
	case PIC_ID(0x2240): // PIC12F609
	case PIC_ID(0x2280): // PIC12HV609
	case PIC_ID(0x2260): // PIC16F610
	case PIC_ID(0x22a0): // PIC16HV610
		printf("  BODEN1: %c\n", (array[7] & 0x0200) ? '1' : '0');
		printf("  BODEN0: %c\n", (array[7] & 0x0100) ? '1' : '0');
		printf("  IOSCFS: %c\n", (array[7] & 0x0080) ? '1' : '0');
		printf("      CP: %c\n", (array[7] & 0x0040) ? '1' : '0');
		printf("   MCLRE: %c\n", (array[7] & 0x0020) ? '1' : '0');
		printf("   PWRTE: %c\n", (array[7] & 0x0010) ? '1' : '0');
		printf("    WDTE: %c\n", (array[7] & 0x0008) ? '1' : '0');
		printf("   FOSC2: %c\n", (array[7] & 0x0004) ? '1' : '0');
		printf("   FOSC1: %c\n", (array[7] & 0x0002) ? '1' : '0');
		printf("   FOSC0: %c\n", (array[7] & 0x0001) ? '1' : '0');
		break;
	case PIC_ID(0x1380): // PIC16F917
	case PIC_ID(0x13a0): // PIC16F916
	case PIC_ID(0x13c0): // PIC16F914
	case PIC_ID(0x13e0): // PIC16F913
	case PIC_ID(0x1460): // PIC16F946
		printf("   DEBUG: %c\n", (array[7] & 0x1000) ? '1' : '0');
		printf("   FCMEN: %c\n", (array[7] & 0x0800) ? '1' : '0');
		printf("    IESO: %c\n", (array[7] & 0x0400) ? '1' : '0');
		printf("  BODEN1: %c\n", (array[7] & 0x0200) ? '1' : '0');
		printf("  BODEN0: %c\n", (array[7] & 0x0100) ? '1' : '0');
		printf("     CPD: %c\n", (array[7] & 0x0080) ? '1' : '0');
		printf("      CP: %c\n", (array[7] & 0x0040) ? '1' : '0');
		printf("   MCLRE: %c\n", (array[7] & 0x0020) ? '1' : '0');
		printf("   PWRTE: %c\n", (array[7] & 0x0010) ? '1' : '0');
		printf("    WDTE: %c\n", (array[7] & 0x0008) ? '1' : '0');
		printf("   FOSC2: %c\n", (array[7] & 0x0004) ? '1' : '0');
		printf("   FOSC1: %c\n", (array[7] & 0x0002) ? '1' : '0');
		printf("   FOSC0: %c\n", (array[7] & 0x0001) ? '1' : '0');
		break;
	case PIC_ID(0x0fa0): // PIC12F635
	case PIC_ID(0x10a0): // PIC16F636/PIC16F639
		printf("    WURE: %c\n", (array[7] & 0x1000) ? '1' : '0');
		printf("   FCMEN: %c\n", (array[7] & 0x0800) ? '1' : '0');
		printf("    IESO: %c\n", (array[7] & 0x0400) ? '1' : '0');
		printf("  BOREN1: %c\n", (array[7] & 0x0200) ? '1' : '0');
		printf("  BOREN0: %c\n", (array[7] & 0x0100) ? '1' : '0');
		printf("     CPD: %c\n", (array[7] & 0x0080) ? '1' : '0');
		printf("      CP: %c\n", (array[7] & 0x0040) ? '1' : '0');
		printf("   MCLRE: %c\n", (array[7] & 0x0020) ? '1' : '0');
		printf("   PWRTE: %c\n", (array[7] & 0x0010) ? '1' : '0');
		printf("    WDTE: %c\n", (array[7] & 0x0008) ? '1' : '0');
		printf("   FOSC2: %c\n", (array[7] & 0x0004) ? '1' : '0');
		printf("   FOSC1: %c\n", (array[7] & 0x0002) ? '1' : '0');
		printf("   FOSC0: %c\n", (array[7] & 0x0001) ? '1' : '0');
		break;
	case PIC_ID(0x0460): // PIC12F683
	case PIC_ID(0x1420): // PIC16F631
	case PIC_ID(0x1440): // PIC16F677
	case PIC_ID(0x1080): // PIC16F684
	case PIC_ID(0x04a0): // PIC16F685
	case PIC_ID(0x1320): // PIC16F687
	case PIC_ID(0x1180): // PIC16F688
	case PIC_ID(0x1340): // PIC16F689
	case PIC_ID(0x1400): // PIC16F690
		printf("   FCMEN: %c\n", (array[7] & 0x0800) ? '1' : '0');
		printf("    IESO: %c\n", (array[7] & 0x0400) ? '1' : '0');
		printf("  BOREN1: %c\n", (array[7] & 0x0200) ? '1' : '0');
		printf("  BOREN0: %c\n", (array[7] & 0x0100) ? '1' : '0');
		printf("     CPD: %c\n", (array[7] & 0x0080) ? '1' : '0');
		printf("      CP: %c\n", (array[7] & 0x0040) ? '1' : '0');
		printf("   MCLRE: %c\n", (array[7] & 0x0020) ? '1' : '0');
		printf("   PWRTE: %c\n", (array[7] & 0x0010) ? '1' : '0');
		printf("    WDTE: %c\n", (array[7] & 0x0008) ? '1' : '0');
		printf("   FOSC2: %c\n", (array[7] & 0x0004) ? '1' : '0');
		printf("   FOSC1: %c\n", (array[7] & 0x0002) ? '1' : '0');
		printf("   FOSC0: %c\n", (array[7] & 0x0001) ? '1' : '0');
		break;
	case PIC_ID(0x0560): // PIC16F84A
		printf("      DP: %c\n", (array[7] & 0x0080) ? '1' : '0');
		printf("      CP: %c\n", (array[7] & 0x0010) ? '1' : '0');
		printf("  PWTREN: %c\n", (array[7] & 0x0008) ? '1' : '0');
		printf("   WDTEN: %c\n", (array[7] & 0x0004) ? '1' : '0');
		printf("   FOSC1: %c\n", (array[7] & 0x0002) ? '1' : '0');
		printf("   FOSC0: %c\n", (array[7] & 0x0001) ? '1' : '0');
		break;
	case PIC_ID(0x0f80): // PIC12F629
	case PIC_ID(0x0fc0): // PIC12F675
	case PIC_ID(0x1fc0): // PIC16F630
	case PIC_ID(0x1fe0): // PIC16F676
		printf("     BG1: %c\n", (array[7] & 0x2000) ? '1' : '0');
		printf("     BG0: %c\n", (array[7] & 0x1000) ? '1' : '0');
		printf("     CPD: %c\n", (array[7] & 0x0100) ? '1' : '0');
		printf("      CP: %c\n", (array[7] & 0x0080) ? '1' : '0');
		printf("   BODEN: %c\n", (array[7] & 0x0040) ? '1' : '0');
		printf("   MCLRE: %c\n", (array[7] & 0x0020) ? '1' : '0');
		printf("   PWRTE: %c\n", (array[7] & 0x0010) ? '1' : '0');
		printf("    WDTE: %c\n", (array[7] & 0x0008) ? '1' : '0');
		printf("   FOSC2: %c\n", (array[7] & 0x0004) ? '1' : '0');
		printf("   FOSC1: %c\n", (array[7] & 0x0002) ? '1' : '0');
		printf("   FOSC0: %c\n", (array[7] & 0x0001) ? '1' : '0');
		break;
	case PIC_ID(0x1220): // PIC16HV785
	case PIC_ID(0x1200): // PIC16F785
		printf("   FCMEN: %c\n", (array[7] & 0x0800) ? '1' : '0');
		printf("    IESO: %c\n", (array[7] & 0x0400) ? '1' : '0');
		printf("  BODEN1: %c\n", (array[7] & 0x0200) ? '1' : '0');
		printf("  BODEN0: %c\n", (array[7] & 0x0100) ? '1' : '0');
		printf("     CPD: %c\n", (array[7] & 0x0080) ? '1' : '0');
		printf("      CP: %c\n", (array[7] & 0x0040) ? '1' : '0');
		printf("   MCLRE: %c\n", (array[7] & 0x0020) ? '1' : '0');
		printf("   PWRTE: %c\n", (array[7] & 0x0010) ? '1' : '0');
		printf("    WDTE: %c\n", (array[7] & 0x0008) ? '1' : '0');
		printf("   FOSC2: %c\n", (array[7] & 0x0004) ? '1' : '0');
		printf("   FOSC1: %c\n", (array[7] & 0x0002) ? '1' : '0');
		printf("   FOSC0: %c\n", (array[7] & 0x0001) ? '1' : '0');
		break;
	case PIC_ID(0x0ba0): // PIC16F737
	case PIC_ID(0x0be0): // PIC16F747
	case PIC_ID(0x0ea0): // PIC16F767
	case PIC_ID(0x0de0): // PIC16F777
		printf("      CP: %c\n", (array[7] & 0x2000) ? '1' : '0');
		printf("   CCPMX: %c\n", (array[7] & 0x1000) ? '1' : '0');
		printf("    RESV: %c\n", (array[7] & 0x0800) ? '1' : '0');
		printf("   BORV1: %c\n", (array[7] & 0x0100) ? '1' : '0');
		printf("   BORV0: %c\n", (array[7] & 0x0080) ? '1' : '0');
		printf("   BOREN: %c\n", (array[7] & 0x0040) ? '1' : '0');
		printf("   MCLRE: %c\n", (array[7] & 0x0020) ? '1' : '0');
		printf("   FOSC2: %c\n", (array[7] & 0x0010) ? '1' : '0');
		printf("  PWTREN: %c\n", (array[7] & 0x0008) ? '1' : '0');
		printf("   WDTEN: %c\n", (array[7] & 0x0004) ? '1' : '0');
		printf("   FOSC1: %c\n", (array[7] & 0x0002) ? '1' : '0');
		printf("   FOSC0: %c\n", (array[7] & 0x0001) ? '1' : '0');
		break;
	case PIC_ID(0x04c0): // PIC16F818
	case PIC_ID(0x04e0): // PIC16F819
	case PIC_ID(0x0720): // PIC16F87
	case PIC_ID(0x0760): // PIC16F88
		printf("      CP: %c\n", (array[7] & 0x2000) ? '1' : '0');
		printf("   CCPMX: %c\n", (array[7] & 0x1000) ? '1' : '0');
		printf("   DEBUG: %c\n", (array[7] & 0x0800) ? '1' : '0');
		printf("    WRT1: %c\n", (array[7] & 0x0400) ? '1' : '0');
		printf("    WRT0: %c\n", (array[7] & 0x0200) ? '1' : '0');
		printf("     CPD: %c\n", (array[7] & 0x0100) ? '1' : '0');
		printf("     LVP: %c\n", (array[7] & 0x0080) ? '1' : '0');
		printf("   BOREN: %c\n", (array[7] & 0x0040) ? '1' : '0');
		printf("   MCLRE: %c\n", (array[7] & 0x0020) ? '1' : '0');
		printf("   FOSC2: %c\n", (array[7] & 0x0010) ? '1' : '0');
		printf("  PWRTEN: %c\n", (array[7] & 0x0008) ? '1' : '0');
		printf("   WDTEN: %c\n", (array[7] & 0x0004) ? '1' : '0');
		printf("   FOSC1: %c\n", (array[7] & 0x0002) ? '1' : '0');
		printf("   FOSC0: %c\n", (array[7] & 0x0001) ? '1' : '0');
		break;
	case PIC_ID(0x1140): // PIC16F716
		printf("      CP: %c\n", (array[7] & 0x2000) ? '1' : '0');
		printf("    BORV: %c\n", (array[7] & 0x0080) ? '1' : '0');
		printf("   BOREN: %c\n", (array[7] & 0x0040) ? '1' : '0');
		printf("   PWRTE: %c\n", (array[7] & 0x0008) ? '1' : '0');
		printf("    WDTE: %c\n", (array[7] & 0x0004) ? '1' : '0');
		printf("   FOSC1: %c\n", (array[7] & 0x0002) ? '1' : '0');
		printf("   FOSC0: %c\n", (array[7] & 0x0001) ? '1' : '0');
		break;
	case PIC_ID(0x0600): // PIC16F73
	case PIC_ID(0x0620): // PIC16F74
	case PIC_ID(0x0640): // PIC16F76
	case PIC_ID(0x0660): // PIC16F77
		printf("   BODEN: %c\n", (array[7] & 0x0040) ? '1' : '0');
		printf("     CP0: %c\n", (array[7] & 0x0010) ? '1' : '0');
		printf("   PWRTE: %c\n", (array[7] & 0x0008) ? '1' : '0');
		printf("    WDTE: %c\n", (array[7] & 0x0004) ? '1' : '0');
		printf("   FOSC1: %c\n", (array[7] & 0x0002) ? '1' : '0');
		printf("   FOSC0: %c\n", (array[7] & 0x0001) ? '1' : '0');
		break;
	case PIC_ID(0x0e40): // PIC16F873A
	case PIC_ID(0x0e60): // PIC16F874A
	case PIC_ID(0x0e00): // PIC16F876A
	case PIC_ID(0x0e20): // PIC16F877A
		printf("      CP: %c\n", (array[7] & 0x2000) ? '1' : '0');
		printf("   DEBUG: %c\n", (array[7] & 0x0800) ? '1' : '0');
		printf("    WRT1: %c\n", (array[7] & 0x0400) ? '1' : '0');
		printf("    WRT0: %c\n", (array[7] & 0x0200) ? '1' : '0');
		printf("     CPD: %c\n", (array[7] & 0x0100) ? '1' : '0');
		printf("     LVP: %c\n", (array[7] & 0x0080) ? '1' : '0');
		printf("   BOREN: %c\n", (array[7] & 0x0040) ? '1' : '0');
		printf("  PWRTEN: %c\n", (array[7] & 0x0008) ? '1' : '0');
		printf("   WDTEN: %c\n", (array[7] & 0x0003) ? '1' : '0');
		printf("   FOSC1: %c\n", (array[7] & 0x0002) ? '1' : '0');
		printf("   FOSC0: %c\n", (array[7] & 0x0001) ? '1' : '0');
		break;
	case PIC_ID(0x07a0): // PIC16F627
	case PIC_ID(0x07c0): // PIC16F628
		printf("     CP1: %c\n", (array[7] & 0x0800) ? '1' : '0');
		printf("     CP0: %c\n", (array[7] & 0x0400) ? '1' : '0');
		printf("     CPD: %c\n", (array[7] & 0x0100) ? '1' : '0');
		printf("     LVP: %c\n", (array[7] & 0x0080) ? '1' : '0');
		printf("   BOREN: %c\n", (array[7] & 0x0040) ? '1' : '0');
		printf("   MCLRE: %c\n", (array[7] & 0x0020) ? '1' : '0');
		printf("   FOSC2: %c\n", (array[7] & 0x0010) ? '1' : '0');
		printf("  PWRTEN: %c\n", (array[7] & 0x0008) ? '1' : '0');
		printf("   WDTEN: %c\n", (array[7] & 0x0004) ? '1' : '0');
		printf("   FOSC1: %c\n", (array[7] & 0x0002) ? '1' : '0');
		printf("   FOSC0: %c\n", (array[7] & 0x0001) ? '1' : '0');
		break;
	case PIC_ID(0x0d00): // PIC16F870
	case PIC_ID(0x0d20): // PIC16F871
	case PIC_ID(0x08e0): // PIC16F872
	case PIC_ID(0x0960): // PIC16F873
	case PIC_ID(0x0920): // PIC16F874
	case PIC_ID(0x09e0): // PIC16F876
	case PIC_ID(0x09a0): // PIC16F877
		printf("    RESV: %c\n", (array[7] & 0x0800) ? '1' : '0');
		printf("     WRT: %c\n", (array[7] & 0x0200) ? '1' : '0');
		printf("     CPD: %c\n", (array[7] & 0x0100) ? '1' : '0');
		printf("     LVP: %c\n", (array[7] & 0x0080) ? '1' : '0');
		printf("   BODEN: %c\n", (array[7] & 0x0040) ? '1' : '0');
		printf("     CP1: %c\n", (array[7] & 0x0020) ? '1' : '0');
		printf("     CP0: %c\n", (array[7] & 0x0010) ? '1' : '0');
		printf("   PWRTE: %c\n", (array[7] & 0x0008) ? '1' : '0');
		printf("    WDTE: %c\n", (array[7] & 0x0004) ? '1' : '0');
		printf("   FOSC1: %c\n", (array[7] & 0x0002) ? '1' : '0');
		printf("   FOSC0: %c\n", (array[7] & 0x0001) ? '1' : '0');
		break;
	}

	putchar('\n');
}

static void encode_inhx8m(FILE *fp, const unsigned char *ptr, size_t addr, size_t size)
{
	int i, sum;
	unsigned char tmp[4];

	tmp[0] = (unsigned char)size;
	tmp[1] = (unsigned char)(addr >> 8);
	tmp[2] = (unsigned char)addr;
	tmp[3] = 0;

	sum = 0;

	putc(':', fp);
	for (i = 0; i < 4; i++) {
		fprintf(fp, "%02X", tmp[i]);
		sum += tmp[i];
	}
	for (i = 0; i < (int)size; i++) {
		fprintf(fp, "%02X", ptr[i]);
		sum += ptr[i];
	}
	fprintf(fp, "%02X\n", (unsigned char)-sum);
}

static void encode_inhx16(FILE *fp, const unsigned short *ptr, size_t addr, size_t size)
{
	int i, sum;
	unsigned char tmp[4];

	tmp[0] = (unsigned char)size;
	tmp[1] = (unsigned char)(addr >> 8);
	tmp[2] = (unsigned char)addr;
	tmp[3] = 0;

	sum = 0;

	putc(':', fp);
	for (i = 0; i < 4; i++) {
		fprintf(fp, "%02X", tmp[i]);
		sum += tmp[i];
	}
	for (i = 0; i < (int)size; i++) {
		fprintf(fp, "%04X", ptr[i]);
		sum += (unsigned char)ptr[i];
		sum += (unsigned char)(ptr[i] >> 8);
	}
	fprintf(fp, "%02X\n", (unsigned char)-sum);
}

void XP::output_binary_le(FILE *fp, const unsigned short *ptr, size_t addr, size_t size)
{
	if (!fp) {
		fp = stdout;
	}
	unsigned long i;
	for (i = 0; i < size; i++) {
		putc((unsigned char)ptr[i], fp);
		putc((unsigned char)(ptr[i] >> 8), fp);
		addr++;
	}
}

void XP::output_binary_be(FILE *fp, const unsigned short *ptr, size_t addr, size_t size)
{
	if (!fp) {
		fp = stdout;
	}
	unsigned long i;
	for (i = 0; i < size; i++) {
		putc((unsigned char)(ptr[i] >> 8), fp);
		putc((unsigned char)ptr[i], fp);
		addr++;
	}
}

void XP::output_mot(FILE *fp, const unsigned short *ptr, size_t addr, size_t size)
{
}

void XP::output_inhx8m(FILE *fp, const unsigned short *ptr, size_t addr, size_t size)
{
	if (!fp) {
		fp = stdout;
	}
	size_t i, len;
	const int STEP = 8;
	unsigned char tmp[STEP * 2];
	while (size > 0) {
		len = size < STEP ? size : STEP;
		{
			for (i = 0; i < len; i++) {
				tmp[i * 2 + 0] = (unsigned char)(ptr[i]);
				tmp[i * 2 + 1] = (unsigned char)(ptr[i] >> 8);
			}
			ptr += len;
		}
		encode_inhx8m(fp, tmp, addr * 2, len * 2);
		addr += len;
		size -= len;
	}
	fprintf(fp, ":00000001ff\n");
}

void XP::input_binary_le(FILE *fp)
{
	if (!fp) {
		fp = stdin;
	}

	size_t l;
	unsigned short tmp[65536];

	tmp[0] = 1;
	bool byteswap = !*(unsigned char *)tmp;		// bigendian ?

	l = fread((char *)tmp, 1, sizeof(tmp), fp);
	if (l < 0) return;

	unsigned long addr = 0;
	unsigned long size = (unsigned long)_option.device.progmemsize;
	_connection.picLoadProgram(tmp, addr, size, byteswap);
}

void XP::input_binary_be(FILE *fp)
{
	if (!fp) {
		fp = stdin;
	}

	unsigned long l;
	unsigned short tmp[65536];

	tmp[0] = 1;
	bool byteswap = !*(unsigned char *)tmp;		// bigendian ?

	l = (unsigned long)fread((char *)tmp, 1, sizeof(tmp), fp);
	if (l < 0) return;

	unsigned long addr = 0;
	unsigned long size = (unsigned long)_option.device.progmemsize;
	_connection.picLoadProgram(tmp, addr, size, !byteswap);
}


static int chrhex(int c)
{
	if (isdigit(c))		return c - '0';
	else if (isxdigit(c))	return toupper(c) - 'A' + 10;
	return -1;
}

bool XP::pic_bulk_write(std::list<bin_t> const *bin, write_error_info_t *errorinfo)
{
	int a = 0;
	int count = 0;
	_connection.picStart();
	for (std::list<bin_t>::const_iterator it = bin->begin(); it != bin->end(); it++) {
		int n = (int)(it->data.size() / 2);
		int addr = it->addr / 2;
		printf("%04X :", addr);
		std::vector<unsigned short> buffer;
		for (int i = 0; i < n; i++) {
			unsigned short c = (it->data[i * 2 + 1] << 8) | it->data[i * 2 + 0];
			buffer.push_back(c);
			printf(" %04X", c);
		}
		bool success = _connection.picBulkLoadProgram(&buffer[0], buffer.size(), errorinfo);
		if (!success) {
			errorinfo->address += addr;
			putchar('\n');
			return false;
		}
		putchar('\n');
		count += n;
	}
	_connection.picEndProgram();
	printf("%d words written\n", count);
	return true;
}

static bool isbigendian()
{
	unsigned short t = 1;
	return !*(unsigned char *)&t;
}

bool XP::avr_write(unsigned short addr, int size, unsigned short const *data, write_error_info_t *errorinfo)
{
	if (size > 64) {
		size = 64;
	}
	bool nl = true;
	for (int i = 0; i < size; i++) {
		if (i % 16 == 0) {
			if (!nl) {
				putchar('\n');
				nl = true;
			}
			printf("%04X :", addr + i);
		}
		printf(" %04X", data[i]);
		nl = false;
	}
	if (!nl) {
		putchar('\n');
	}
	_connection.avrLoadProgram(addr, size, data);

	//

	std::vector<unsigned short> tmp(size);
	_connection.avrReadProgram(addr, size, &tmp[0]);
	for (int i = 0; i < size; i++) {
		if (tmp[i] != data[i]) {
			errorinfo->address = addr + i;
			errorinfo->write = data[i];
			errorinfo->read = tmp[i];
			return false;
		}
	}

	return true;
}

bool XP::avr_write(unsigned short addr, int size, unsigned char const *data, write_error_info_t *errorinfo)
{
	size /= 2;
	if (size > 64) {
		size = 64;
	}
	const int pagesize = 64;
	int offset = 0;
	while (offset < size) {
		unsigned short tmp[pagesize];
		int len = size - offset;
		if (len > pagesize) {
			len = pagesize;
		}
		for (int i = 0; i < len; i++) {
			tmp[i] = (data[(offset + i) * 2 + 1] << 8) | data[(offset + i) * 2 + 0];
		}
		if (!avr_write(addr + offset, len, tmp, errorinfo)) {
			return false;
		}
		offset += len;
	}
	return true;
}







struct HexParserInfo {
	unsigned short bitmap[65536 / 16];
	unsigned char data[65536];
	void (*output)(unsigned long addr, const unsigned char *ptr, int len, void *cookie);
	void *cookie;
};

bool parse_motorola_s_record(struct HexParserInfo *info, const char *line, int linenumber)
{
	int c, i, n, len, type, bytes, length;
	unsigned char *ptr, tmp[1000];
	unsigned long offset;
	unsigned char sum;

	len = (int)strlen(line);

	if (len > 4 && line[0] == 'S') {		// motorola s record format

		type = line[1];
		sscanf(line + 2, "%02x", &bytes);
		if (len < 4 + bytes * 2) {
			return false;
		} else {
			sum = bytes;
			for (i = 0; i < bytes; i++) {
				sscanf(line + 4 + i * 2, "%02x", &c);
				tmp[i] = c;
				sum += c;
			}
			if (sum != 0xff) {
				printf("warning: checksum incorrect, line %u\n", linenumber);
			}
			offset = (tmp[0] << 8) | tmp[1];
			n = 2;
			switch (type) {
			case '3':
				offset = (offset << 8) | tmp[n++];
				// fallthru
			case '2':
				offset = (offset << 8) | tmp[n++];
				// fallthru
			case '1':
				if (n + 1 < bytes) {
					ptr = tmp + n;
					length = bytes - n - 1;
					info->output(offset, ptr, length, info->cookie);
				} else {
					return false;
				}
				break;
			}
		}
	} else {
		while (isspace(*line)) line++;
		if (*line) {
			return false;
		}
	}

	return true;
}

bool parse_motorola_s_record_file(struct HexParserInfo *info, FILE *fp)
{
	int i, c, linenumber;
	char tmp[1000];

	linenumber = 0;

	while (!feof(fp)) {
		i = 0;
		while ((c = getc(fp)) != EOF) {
			if (c == '\n') break;
			if (i < sizeof(tmp) - 1) {
				tmp[i++] = c;
			}
		}
		tmp[i] = 0;
		if (!parse_motorola_s_record(info, tmp, linenumber + 1)) {
			return false;
		}
		linenumber++;
	}

	return true;
}

bool XP::h8_transfer_boot_program(const unsigned char *loader, size_t length)
{
	unsigned char tmp[256];

	{
		bool success = false;
		for (int i = 1; i < 20; i++) {
			for (int j = 0; j < 10; j++) {
				tmp[0] = 0;
				_connection.Write(tmp, 1);
			}
			if (_connection.Read(tmp, 1, 1)) {
				if (tmp[0] == 0) {
					tmp[0] = 0x55;
					_connection.Write(tmp, 1);
					_connection.Read(tmp, 1, 1000);
					if (tmp[0] == 0xaa) {
						success = true;
						i = 0;
					}
					break;
				}
			}
		}
		if (!success) {
			goto failed;
		}
	}

	fprintf(stderr, "Target device found.\n");

	tmp[0] = (unsigned char)(length >> 8);
	tmp[1] = (unsigned char)length;
	_connection.Write(tmp, 2);
	_connection.Read(tmp, 2, 1000);
	{
		unsigned char h = tmp[0];
		unsigned char l = tmp[1];
		if (length != ((h << 8) | l)) {
			goto failed;
		}
	}

	for (int i = 0; (size_t)i < length; i += 256) {
		size_t n = length - (size_t)i;
		if (n > 256) n = 256;
		_connection.Write(loader + i, n);
		if (!_connection.Read(tmp, n, 1000)) {
			goto failed;
		}
		if (memcmp(loader + i, tmp, n) != 0) {
			goto failed;
		}
	}

	_connection.Read(tmp, 1, 1000);
	if (tmp[0] != 0xaa) {
		goto failed;
	}


	fprintf(stderr, "Boot program transfered.\n");
	return true;

failed:
	fprintf(stderr, "Boot program transfer failed.\n");
	return false;
}


void h8_output_callback(unsigned long addr, const unsigned char *ptr, int len, void *cookie)
{
	int i;
	struct H8DeviceInfo *info = (struct H8DeviceInfo *)cookie;
	for (i = 0; i < len; i++) {
		if (addr < info->flashdata.size()) {
			info->flashdata[addr] = ptr[i];
			info->flashmap[addr >> 11] |= 1 << ((addr >> 8) & 7);
		}
		addr++;
	}
}

bool XP::receive_h8_3069_response(unsigned char *dst, int timeout)
{
	int len = 0;
	bool success = false;
	if (!_connection.Read(dst, 1, 10000)) {
		fprintf(stderr, "no respond.\n");
			return false;
	}
	len++;
	if (dst[0] == 0x06) {	// ack
		fprintf(stderr, " %02x (ack)\n", dst[0]);
		return true;
	}
	if (dst[0] & 0x80) {
		if (_connection.Read(dst + 1, 1, 1000)) {
			len++;
		}
	} else {
		if (!_connection.Read(dst + 1, 1, 1000)) {
			fprintf(stderr, "no respond.\n");
		} else {
			fprintf(stderr, "respond:\n", dst[0]);
			len = dst[1];
			if (_connection.Read(dst + 2, len + 1, 1000)) {
				len += 3;
				success = true;
			} else {
				len = 2;
			}
		}
	}
	int i;
	unsigned char sum = 0;
	for (i = 0; i < len; i++) {
		if ((i & 15) == 0) {
			if (i > 0) {
				putc('\n', stderr);
			}
		}
		fprintf(stderr, " %02x", dst[i]);
		sum += dst[i];
	}
	if (i > 0) {
		putc('\n', stderr);
	}
	if (success) {
		fprintf(stderr, "checksum: %02x (%s)\n", sum, sum == 0 ? "ok" : "ng");
		return sum == 0;
	}
	fprintf(stderr, "error: ");
	if (dst[0] & 0x80) {
		if (dst[0] == 0x80) {
			fprintf(stderr, "command error.\n");
		} else {
			switch (dst[1]) {
			case 0x11:
				fprintf(stderr, "checksum incorrect.\n");
				break;
			case 0x22:
				fprintf(stderr, "clock mode incorrect.\n");
				break;
			case 0x24:
				fprintf(stderr, "invalid bit rate.\n");
				break;
			case 0x25:
				fprintf(stderr, "invalid input frequency.\n");
				break;
			case 0x26:
				fprintf(stderr, "invalid multiplication ratio.\n");
				break;
			case 0x27:
				fprintf(stderr, "invalid operationg frequency.\n");
				break;
			case 0x51:
				fprintf(stderr, "erase failed.\n");
				break;
			}
		}
	}
	return false;
}

void set_checksum(unsigned char *ptr, int len)
{
	int i;
	unsigned char sum = 0;
	for (i = 0; i < len; i++) {
		sum += ptr[i];
	}
	ptr[i] = -sum;
}

bool XP::write_h8_3069_128bytes(unsigned long addr, unsigned char const *src)
{
	int i, n;
	unsigned char sum;
	unsigned char tmp[5 + 128 + 1];

	n = 0xff;
	for (i = 0; i < 128; i++) {
		n &= src[i];
	}
	if (n == 0xff) {
		return true;
	}

	// select user mat
	fprintf(stderr, "change erase/write mode.\n");
	n = 1;
	tmp[0] = 0x43;
	_connection.Write(tmp, n);
	if (!receive_h8_3069_response(tmp, 10000)) {
		return false;
	}
	// write
	fprintf(stderr, "writing (0x%04x)\n", addr);
	n = 5 + 128;
	tmp[0] = 0x50;
	tmp[1] = addr >> 24;
	tmp[2] = addr >> 16;
	tmp[3] = addr >> 8;
	tmp[4] = addr;
	memcpy(tmp + 5, src, 128);
	set_checksum(tmp, n++);
	_connection.Write(tmp, n);
	if (!receive_h8_3069_response(tmp)) {
		return false;
	}
	// end write mode
	n = 5;
	tmp[0] = 0x50;
	tmp[1] = 0xff;
	tmp[2] = 0xff;
	tmp[3] = 0xff;
	tmp[4] = 0xff;
	set_checksum(tmp, n++);
	_connection.Write(tmp, n);
	if (!receive_h8_3069_response(tmp)) {
		return false;
	}
	// read
	fprintf(stderr, "reading (0x%04x)\n", addr);
	n = 11;
	tmp[0] = 0x52;
	tmp[1] = n - 2;
	tmp[2] = 1; // user mat
	tmp[3] = addr >> 24;
	tmp[4] = addr >> 16;
	tmp[5] = addr >> 8;
	tmp[6] = addr;
	tmp[7] = 0;
	tmp[8] = 0;
	tmp[9] = 0;
	tmp[10] = 128;
	set_checksum(tmp, n++);
	_connection.Write(tmp, n);
	if (!_connection.Read(tmp, 5, 1000)) {
		fprintf(stderr, "no respond.\n");
		return false;
	}
	if (memcmp(tmp, "\x52\x00\x00\x00\x80", 5) != 0) {
		fprintf(stderr, "reading failed.\n");
		return false;
	}
	if (!_connection.Read(tmp + 5, 128 + 1, 1000)) {
		fprintf(stderr, "no respond.\n");
		return false;
	}
	sum = 0;
	for (i = 0; i < 5 + 128 + 1; i++) {
		sum += tmp[i];
	}
	if (sum != 0) {
		fprintf(stderr, "verification failed.\n");
		return false;
	}
	for (i = 0; i < 128; i++) {
		if ((i & 15) == 0) {
			if (i > 0) {
				putchar('\n');
			}
			printf("%04x :", addr + i);
		}
		printf(" %02x", tmp[5 + i]);
	}
	putchar('\n');

	return true;
}

void XP::input_mot_write(FILE *fp)
{
	int i, j, n, addr;
	struct HexParserInfo srinfo;

	n = _option.device.progmemsize / 256;
	_option.device.h8.flashmap.resize(n);
	memset(&_option.device.h8.flashmap[0], 0, n);

	n *= 256;
	_option.device.h8.flashdata.resize(n);
	memset(&_option.device.h8.flashdata[0], 0xff, n);

	srinfo.output = h8_output_callback;
	srinfo.cookie = &_option.device.h8;

	if (!parse_motorola_s_record_file(&srinfo, fp)) {
		fprintf(stderr, "Invalid file format.\n");
		return;
	}

	// transfer boot program

	if (_option.device.id == H8_ID(0x3069)) {

//		int freq = 25000000;
		int freq = _option.target_clock;
		int bps = 9600;

		bps /= 100;
//		freq = (freq + 5000) / 10000;
		freq = freq / 10000;

		unsigned char tmp[260];

		fprintf(stderr, "attempt entering boot mode.\n");
		{
			bool success = false;
			for (i = 1; i < 10; i++) {
				for (j = 0; j < 3; j++) {
					tmp[0] = 0;
					_connection.Write(tmp, 1);
				}
				if (_connection.Read(tmp, 1, 1)) {
					fprintf(stderr, "respond1: 0x%02x\n", tmp[0]);
					if (tmp[0] == 0) {
						tmp[0] = 0x55;
						_connection.Write(tmp, 1);
						_connection.Read(tmp, 1, 1000);
						fprintf(stderr, "respond2: 0x%02x\n", tmp[0]);
						if (tmp[0] == 0xe6) {
							success = true;
							i = 0;
						}
						break;
					}
				}
			}
			if (!success) {
				goto failed;
			}

			unsigned char sum;
			// query send support device
			fprintf(stderr, "send query support device.\n");
			tmp[0] = 0x20;
			_connection.Write(tmp, 1);
			if (!receive_h8_3069_response(tmp)) {
				goto failed;
			}
			if (tmp[0] != 0x30) {
				fprintf(stderr, "invalid response code. required 0x%02x but 0x%02x.\n", 0x30, tmp[0]);
				goto failed;
			}
			fprintf(stderr, "device: ");
			for (i = 0; i < tmp[3]; i++) {
				putc(tmp[i + 4], stderr);
			}
			putc('\n', stderr);
			if (sum != 0) {
				goto failed;
			}
			// send device select
			fprintf(stderr, "send device select.\n");
			memmove(tmp + 2, tmp + 4, 4);
			n = 6;
			tmp[0] = 0x10;
			tmp[1] = n - 2;
			sum = 0;
			set_checksum(tmp, n++);
			_connection.Write(tmp, n);
			if (!receive_h8_3069_response(tmp)) {
				goto failed;
			}
#if 01
			// query writing size
			fprintf(stderr, "query writing size.\n");
			n = 1;
			tmp[0] = 0x27;
			_connection.Write(tmp, n);
			if (!receive_h8_3069_response(tmp)) {
				goto failed;
			}
#endif
			// query clock mode
			fprintf(stderr, "query clock mode.\n");
			n = 1;
			tmp[0] = 0x21;
			_connection.Write(tmp, n);
			if (!receive_h8_3069_response(tmp)) {
				goto failed;
			}
			// query multiplication ratio
			fprintf(stderr, "query multiplication ratio.\n");
			n = 1;
			tmp[0] = 0x22;
			_connection.Write(tmp, n);
			if (!receive_h8_3069_response(tmp)) {
				goto failed;
			}
			// query operation clock
			fprintf(stderr, "query operation clock.\n");
			n = 1;
			tmp[0] = 0x23;
			_connection.Write(tmp, n);
			if (!receive_h8_3069_response(tmp)) {
				goto failed;
			}
			// set clock mode
			fprintf(stderr, "set clock mode.\n");
			n = 3;
			tmp[0] = 0x11;
			tmp[1] = n - 2;
			tmp[2] = 0;
			set_checksum(tmp, n++);
			_connection.Write(tmp, n);
			if (!receive_h8_3069_response(tmp)) {
				goto failed;
			}
			// set bit rate (0x3f)
			fprintf(stderr, "set bit rate.\n");
			n = 9;
			tmp[0] = 0x3f;
			tmp[1] = n - 2;
			tmp[2] = bps >> 8;
			tmp[3] = bps;
			tmp[4] = freq >> 8;
			tmp[5] = freq;
			tmp[6] = 1;
			tmp[7] = 1;
			tmp[8] = 1;
			set_checksum(tmp, n++);
			_connection.Write(tmp, n);
			if (!receive_h8_3069_response(tmp)) {
				goto failed;
			}
			if (tmp[0] != 0x06) {
				fprintf(stderr, "invalid bit rate.\n");
				goto failed;
			}
			n = 1;
			tmp[0] = 0x06;
			_connection.Write(tmp, n);
			if (!receive_h8_3069_response(tmp)) {
				goto failed;
			}
			// query writing size
			fprintf(stderr, "query writing size.\n");
			n = 1;
			tmp[0] = 0x27;
			_connection.Write(tmp, n);
			if (!receive_h8_3069_response(tmp)) {
				goto failed;
			}
			// query status
			fprintf(stderr, "query status.\n");
			n = 1;
			tmp[0] = 0x4f;
			_connection.Write(tmp, n);
			if (!receive_h8_3069_response(tmp)) {
				goto failed;
			}
			// change erase/write mode
			fprintf(stderr, "change erase/write mode.\n");
			n = 1;
			tmp[0] = 0x40;
			_connection.Write(tmp, n);
			if (!receive_h8_3069_response(tmp, 10000)) {
				goto failed;
			}
			//
			fprintf(stderr, "writing...\n");
			n = _option.device.progmemsize / 256;
			for (i = 0; i < n; i++) {
				addr = i * 256;
				if ((_option.device.h8.flashmap[i >> 3] >> (i & 7)) & 1) {
					if (!write_h8_3069_128bytes(addr, &_option.device.h8.flashdata[addr])) {
						goto failed;
					}
					if (!write_h8_3069_128bytes(addr + 0x80, &_option.device.h8.flashdata[addr] + 0x80)) {
						goto failed;
					}
				}
			}
		}

	} else {

		if (!h8_transfer_boot_program(_option.device.h8.bootdata, _option.device.h8.bootsize)) {
			return;
		}

		{ // dummy read
			unsigned char tmp[2];
			if (!_connection.Read(tmp, 2, 1000)) {
				fprintf(stderr, "Target not respond.\n");
				return;
			}
			if ((tmp[0] ^ tmp[1]) != 0xff) {
				fprintf(stderr, "Something wrong.\n");
				return;
			}
			fprintf(stderr, "BRR: 0x%02x\n", tmp[0]);
		}

		// transfer user program

		n = _option.device.progmemsize / 256;
		for (i = 0; i < n; i++) {
			addr = i * 256;
			if ((_option.device.h8.flashmap[i >> 3] >> (i & 7)) & 1) {
				unsigned char tmp[100];

				tmp[0] = OP_SETBINARY256;
				_connection.Write(tmp, 1);
				if (!_connection.Write(&_option.device.h8.flashdata[addr], 256)) {
					goto failed;
				}

				// set flash memory address
				tmp[0] = (unsigned char)((addr >> 28) & 127);
				tmp[1] = (unsigned char)((addr >> 21) & 127);
				tmp[2] = (unsigned char)((addr >> 14) & 127);
				tmp[3] = (unsigned char)((addr >> 7) & 127);
				tmp[4] = (unsigned char)(addr & 127);
				tmp[5] = OP_SETADDRESS;
				if (!_connection.Write(tmp, 6)) {
					goto failed;
				}

				// program
				tmp[0] = OP_PROGRAM;
				if (!_connection.Write(tmp, 1)) {
					goto failed;
				}

				// verify
				for (j = 0; j < 256; j++) {
					if (!_connection.Read(tmp, 1, 1000)) {
						goto failed;
					}
					if ((j & 15) == 0) {
						printf("%04x :", addr + j);
					}
					printf(" %02x", tmp[0]);
					if ((j & 15) == 15) {
						putchar('\n');
					}
				}
			}
		}
	}
failed:;
}

void align(std::list<bin_t> *bin)
{
	std::list<bin_t> newdata;
	for (std::list<bin_t>::const_iterator it = bin->begin(); it != bin->end(); it++) {
		bool append = false;
		if (!newdata.empty() && newdata.back().addr + newdata.back().data.size() == it->addr) {
			append = true;
		}
		if (!append) {
			newdata.insert(newdata.end(), bin_t());
			newdata.back().addr = it->addr;
		}
		newdata.back().data.insert(newdata.back().data.end(), it->data.begin(), it->data.end());
	}
	bin->clear();
	for (std::list<bin_t>::const_iterator it = newdata.begin(); it != newdata.end(); it++) {
		unsigned short addr = it->addr;
		unsigned char const *data = &it->data[0];
		size_t size = it->data.size();
		size_t end = addr + size;
		while (addr < end) {
			char tmp[16];
			int a = addr & 0x000f;
			int e = addr + 16 - a;
			if (e > end) {
				e = end;
			}
			size_t n = e - addr;
			memset(tmp, 0xff, a);
			memcpy(tmp + a, data, n);
			bin->insert(bin->end(), bin_t());
			bin->back().addr = addr - a;
			bin->back().data.insert(bin->back().data.end(), tmp, tmp + a + n);
			addr += n;
			data += n;
		}
	}
}

void XP::input_inhx8m_write(FILE *fp)
{
	if (!fp) {
		fp = stdin;
	}

	if (DEVICE_TYPE(_option.device.id) == DEVICE_TYPE_AVR) {
		avr_program_mode(true);
		avr_chip_erase();
		avr_program_mode(true);
	}

	int c, pos, count;
	unsigned char tmp[100];
	unsigned char buffer[65536 * 2];

	std::list<bin_t> bin;

	bool bigendian = isbigendian();

	count = 0;
	memset(buffer, 255, sizeof(buffer));

	pos = -1;
	while (true) {

		c = getc(fp);
		if (c == EOF) break;

		if (c == ':') {
			pos = 0;
		} else if (isxdigit(c)) {
			if (pos >= 0 && pos < sizeof(tmp) * 2) {
				c = chrhex(c);
				if (pos & 1) tmp[pos / 2] |= c;
				else         tmp[pos / 2] = c << 4;
				pos++;
			}
		} else {
			if ((c == '\n' || c == '\r') && pos > 8) {
				int i, n;
				unsigned char sum;
				n = pos / 2;
				if (n == tmp[0] + 5) {
					for (sum = i = 0; i < n; sum += tmp[i++]);
					if (sum) { // incorrect check sum
						return;
					}
					unsigned short addr = ((tmp[1] << 8) | tmp[2]);
					if (tmp[3] == 0 || tmp[3] == 1) {
						n = tmp[0];
						if (n > 0) {
							for (i = 0; i < n; i++) {
								int a = i;
								if (bigendian) a ^= 1;
								buffer[a] = tmp[i + 4];
							}
							count += n;
							bin.push_back(bin_t());
							bin.back().addr = addr;
							bin.back().data.assign(buffer, buffer + n);
						}
					}
				}
			}
			pos = -1;
		}
	}

	align(&bin);

	if (count > 0) {
		bool success = true;
		write_error_info_t errorinfo;
		if (DEVICE_TYPE(_option.device.id) == DEVICE_TYPE_PIC) {
			success = pic_bulk_write(&bin, &errorinfo);
		} else if (DEVICE_TYPE(_option.device.id) == DEVICE_TYPE_AVR) {
			switch (_option.device.id) {
			case AVR_ID(0x910a):
			case AVR_ID(0x9205):
			case AVR_ID(0x930a):
			case AVR_ID(0x9406):
			case AVR_ID(0x920a):
			case AVR_ID(0x930f):
			case AVR_ID(0x940B):
			case AVR_ID(0x950f):
				{ // y[WPʂŏ
					_connection.SetChipMode(0x81);
					for (std::list<bin_t>::const_iterator it = bin.begin(); it != bin.end(); it++) {
						if (!avr_write(it->addr / 2, (int)it->data.size(), &it->data.at(0), &errorinfo)) {
							success = false;
							break;
						}
					}
				}
				break;
			default:
				{ // [hPʂŏ
					_connection.SetChipMode(0x80);
					for (std::list<bin_t>::const_iterator it = bin.begin(); it != bin.end(); it++) {
						if (!avr_write(it->addr / 2, (int)it->data.size(), &it->data.at(0), &errorinfo)) {
							success = false;
							break;
						}
					}
				}
				break;
			}
		}
		if (!success) {
			printf(
				"Verification failure.\n"
				"    a: %04X\n"
				"    w: %04X\n"
				"    r: %04X\n"
				, errorinfo.address
				, errorinfo.write
				, errorinfo.read
				);
		}
	}

	if (DEVICE_TYPE(_option.device.id) == DEVICE_TYPE_AVR) {
		avr_program_mode(false);
	}
}


void XP::Read()
{
	try {
		int i;
		unsigned long addr, size;

		addr = 0;
		size = (unsigned long)_option.device.progmemsize;

		std::vector<unsigned short> buffer;
		buffer.reserve(size);

		switch (DEVICE_TYPE(_option.device.id)) {
		case DEVICE_TYPE_PIC:
			i = _connection.picReadProgram(&buffer, addr, size);
			if (i) throw "";
			break;
		case DEVICE_TYPE_AVR:
			size /= 2;
			buffer.resize(size);
			{
				avr_program_mode(true);
				avr_program_mode(false);
				avr_program_mode(true);
				int offset = 0;
				while (offset < size) {
					unsigned short tmp[8];
					int len = size - offset;
					if (len > 8) {
						len = 8;
					}
					_connection.avrReadProgram(addr + offset, len, tmp);
					for (int i = 0; i < len; i++) {
						buffer[addr + offset + i] = tmp[i];
					}
					offset += len;
				}
				avr_program_mode(false);
			}
			break;
		}


		FILE *fp = 0;
		//if (!_option.output_console) {
			char const *path = "a.out";
			char const *mode = "w";
			if (!_option.output_file.empty()) {
				path = _option.output_file.c_str();
			}
			fp = fopen(path, mode);
			if (!fp) {
				std::string msg;
				msg = "Unable to open file '";
				msg.append(path);
				msg.append("'");
				throw Exception(msg);
			}
		//}
		switch (_option.format) {
		case FORMAT_SREC:        output_mot(fp, &buffer[0], addr, buffer.size());     break;
		case FORMAT_INHX8M:     output_inhx8m(fp, &buffer[0], addr, buffer.size());     break;
		}
		if (fp) {
			fclose(fp);
		}
	} catch (...) {
		ClosePort();
		throw;
	}
}

void XP::Write()
{
	try {
		FILE *fp = 0;
		//if (!_option.input_console) {
			char const *path = "";
			char const *mode = "r";
			if (!_option.input_file.empty()) {
				path = _option.input_file.c_str();
			}
			fp = fopen(path, mode);
			if (!fp) {
				std::string msg;
				msg = "Unable to open file '";
				msg.append(path);
				msg.append("'");
				throw Exception(msg);
			}
		//}
		switch (_option.format) {
		case FORMAT_SREC:       input_mot_write(fp);     break;
		case FORMAT_INHX8M:    input_inhx8m_write(fp);	break;
		}
	} catch (...) {
		ClosePort();
		throw;
	}
}

void XP::pic_disable_code_protect()
{
	_connection.picDisableCodeProtect();
}

void XP::pic_read_confword()
{
	unsigned short array[8];

	_connection.picReadConfig(array, 8);

	printf("%04x\n", array[7]);
}

void XP::pic_write_confword(unsigned short data)
{
	_connection.picLoadConfig(7, data);
}

void XP::pic_bulk_erase_program_memory()
{
	_connection.picBulkEraseProgramMemory(_option.prompt);
}

void XP::pic_bulk_erase_data_memory()
{
	_connection.picBulkEraseDataMemory();
}

void XP::pic_read_p(unsigned short addr)
{
	if (addr < 0x2000) {
		_connection.picStart();
		
		for (int i = 0; i < addr; i++) {
			_connection.picIncrement();
		}

		unsigned short data = _connection.picReadProgram(false);

		_connection.picEndProgram();

		printf("%04x: %04x\n", addr, data);
	} else {
		unsigned short data = _connection.picReadConfig(addr - 0x2000);
		printf("%04x: %04x\n", addr, data);
	}
}

void XP::pic_write_p(unsigned short addr, unsigned short data)
{
	if (addr < 0x2000) {
		_connection.picStart();
		
		for (int i = 0; i < addr; i++) {
			_connection.picIncrement();
		}

		_connection.picLoadProgram(data);
		_connection.picBeginProgram();

		data = _connection.picReadProgramWord(false);
		printf("%04x: %04x\n", addr, data);

		_connection.picEndProgram();
	} else {
		_connection.picLoadConfig(addr - 0x2000, data);
		printf("%04x: %04x\n", addr, data);
	}
}

void XP::pic_erase_pic12f629()
{
	unsigned short data[8];

	_connection.picReadConfig(data, 8);

	_connection.picStart();
	for (int i = 0; i < 0x3ff; i++) {
		_connection.picIncrement();
	}
	unsigned short osccal = _connection.picReadProgramWord(false);
	_connection.picEndProgram();

	if ((data[6] & 0x3fe0) != 0x0f80) {
		printf("The device is not PIC12F629\n");
		return;
	}

	unsigned short cw = data[7];

	printf("    CW: %04x\n", cw);
	printf("OSCCAL: %04x\n", osccal);

	_connection.picBulkEraseProgramMemory(_option.prompt);
	_connection.picBulkEraseDataMemory();

	cw = (cw & 0x3000) | 0x01ff;

	sleep_ms(100);

	_connection.picStart();
	_connection.picCommand(0x00);
	_connection.picLoad(0x3fff); _connection.picBeginProgram(); _connection.picIncrement();
	_connection.picLoad(0x3fff); _connection.picBeginProgram(); _connection.picIncrement();
	_connection.picLoad(0x3fff); _connection.picBeginProgram(); _connection.picIncrement();
	_connection.picLoad(0x3fff); _connection.picBeginProgram(); _connection.picIncrement();
	_connection.picLoad(0x3fff); _connection.picBeginProgram(); _connection.picIncrement();
	_connection.picLoad(0x3fff); _connection.picBeginProgram(); _connection.picIncrement();
	_connection.picLoad(0x3fff); _connection.picBeginProgram(); _connection.picIncrement();
	_connection.picLoad(cw);     _connection.picBeginProgram();
	_connection.picEndProgram();

	_connection.picStart();
	for (int i = 0; i < 0x3ff; i++) {
		_connection.picIncrement();
	}
	_connection.picLoadProgram(osccal);
	_connection.picBeginProgram();

	_connection.picEndProgram();
}

void XP::avr_program_mode(bool f)
{
	_connection.avrDisable();
	if (f) {
		_connection.avrEnable();
	}
}

void XP::avr_inquire()
{
	unsigned char tmp[3];

	avr_program_mode(true);
	for (int i = 0; i < 3; i++) {
		tmp[i] = _connection.avrReadSignature(i);
	}
	avr_program_mode(false);

	printf("Signature: %02X %02X %02X\n", tmp[0], tmp[1], tmp[2]);

	char const *name = 0;

	switch (AVR_ID((tmp[1] << 8) | tmp[2])) {
	case AVR_ID(0x9101):  name = "AT90S2313";                   break;
	case AVR_ID(0x910a):  name = "ATTINY2313";                  break;
	case AVR_ID(0x9205):  name = "ATMEGA48";                    break;
	case AVR_ID(0x920a):  name = "ATMEGA48PA";                  break;
	case AVR_ID(0x930a):  name = "ATMEGA88";                    break;
	case AVR_ID(0x930f):  name = "ATMEGA88PA";                  break;
	case AVR_ID(0x9406):  name = "ATMEGA168";                   break;
	case AVR_ID(0x940B):  name = "ATMEGA168PA";                 break;
	case AVR_ID(0x950f):  name = "ATMEGA328P";                  break;
	}

	if (!name) {
		name = "Unknown";
	}

	printf("Device: %s\n", name);
}

void XP::avr_chip_erase()
{
	avr_program_mode(true);

	_connection.avrChipErase();

	avr_program_mode(false);
}

void XP::avr_read_fuse()
{
	avr_program_mode(true);

	unsigned char tmp[4];
	_connection.avrReadFuse(tmp);
	printf("         lock bits: %02x\n", tmp[3]);
	printf("extended fuse bits: %02x\n", tmp[2]);
	printf("    fuse high bits: %02x\n", tmp[1]);
	printf("     fuse low bits: %02x\n", tmp[0]);

	avr_program_mode(false);
}

void XP::avr_write_fuse_l(unsigned char c)
{
	avr_program_mode(true);
	_connection.avrWriteFuseL(c);
	avr_program_mode(false);
}

void XP::avr_write_fuse_h(unsigned char c)
{
	avr_program_mode(true);
	_connection.avrWriteFuseH(c);
	avr_program_mode(false);
}

void XP::avr_write_fuse_e(unsigned char c)
{
	avr_program_mode(true);
	_connection.avrWriteFuseE(c);
	avr_program_mode(false);
}






void XP::do_test()
{

}


void XP::FirmwareVersion()
{
	_connection.FirmwareVersion();
}


std::string get_port_path(int n)
{
#ifdef WIN32
	char tmp[100];
	sprintf(tmp, "\\\\.\\COM%u", n + 1);
	return tmp;
#else
	char tmp[100];
	sprintf(tmp, "/dev/ttyS%u", n);
	return tmp;
#endif
}

int XP::main(int argc, char **argv)
{
	enum {
		OPTION_NONE,
		OPTION_TEST,
		OPTION_PIC_BULK_ERASE_P,
		OPTION_PIC_BULK_ERASE_D,
		OPTION_PIC_INQUIRE,
		OPTION_AVR_INQUIRE,
		OPTION_READ,
		OPTION_WRITE,
		OPTION_PORT,
		OPTION_PROGMEMSIZE,
		OPTION_DATAMEMSIZE,
		OPTION_VERSION,
		OPTION_FIRMWARE_VERSION,
		OPTION_DEVICE,
		OPTION_DISABLE_CODE_PROTECT,
		//OPTION_ADDRESS,
		//OPTION_READ_P,
		//OPTION_WRITE_P,
		OPTION_PIC_CONFWORD,
		OPTION_INPUT,
		OPTION_OUTPUT,
		OPTION_INPUT_CONSOLE,
		OPTION_OUTPUT_CONSOLE,
		OPTION_FORMAT,
		OPTION_ERASE_PIC12F629,
		OPTION_AVR_CHIP_ERASE,
		OPTION_AVR_READ_FUSE,
		OPTION_AVR_WRITE_FUSE_L,
		OPTION_AVR_WRITE_FUSE_H,
		OPTION_AVR_WRITE_FUSE_E,
		OPTION_TARGET_CLOCK,
		OPTION_PROMPT,
	};
	static struct {
		char const *text;
		int id;
		bool with_value;
	} optiontable[] = {
		"--test"                    , OPTION_TEST                 , false,
		"--pic-bulk-erase-p"        , OPTION_PIC_BULK_ERASE_P     , false,
		"--pic-bulk-erase-d"        , OPTION_PIC_BULK_ERASE_D     , false,
		"-qp"                       , OPTION_PIC_INQUIRE          , false,
		"-pq"                       , OPTION_PIC_INQUIRE          , false,
		"--pic-inquire"             , OPTION_PIC_INQUIRE          , false,
		"-qa"                       , OPTION_AVR_INQUIRE          , false,
		"-aq"                       , OPTION_AVR_INQUIRE          , false,
		"--avr-inquire"             , OPTION_AVR_INQUIRE          , false,
		"-r"                        , OPTION_READ                 , false,
		"-w"                        , OPTION_WRITE                , false,
		"-p"                        , OPTION_PORT                 , true,
		"-ps"                       , OPTION_PROGMEMSIZE          , true,
		"-ds"                       , OPTION_DATAMEMSIZE          , true,
		"-v"                        , OPTION_VERSION              , false,
		"-vf"                       , OPTION_FIRMWARE_VERSION     , false,
		"-d"                        , OPTION_DEVICE               , true,
		"--disable-code-protect"    , OPTION_DISABLE_CODE_PROTECT , true,
		"--pic-conf-word"           , OPTION_PIC_CONFWORD         , true,
		"-i"                        , OPTION_INPUT                , true,
		"-o"                        , OPTION_OUTPUT               , true,
		"-ic"                       , OPTION_INPUT_CONSOLE        , true,
		"-oc"                       , OPTION_OUTPUT_CONSOLE       , true,
		"-f"                        , OPTION_FORMAT               , true,
		"--erase-pic12f629"         , OPTION_ERASE_PIC12F629      , false,
		"--avr-chip-erase"          , OPTION_AVR_CHIP_ERASE       , false,
		"--avr-read-fuse"           , OPTION_AVR_READ_FUSE        , false,
		"-awfl"                     , OPTION_AVR_WRITE_FUSE_L     , true,
		"-awfh"                     , OPTION_AVR_WRITE_FUSE_H     , true,
		"-awfe"                     , OPTION_AVR_WRITE_FUSE_E     , true,
		"--avr-write-fuse-l"        , OPTION_AVR_WRITE_FUSE_L     , true,
		"--avr-write-fuse-h"        , OPTION_AVR_WRITE_FUSE_H     , true,
		"--avr-write-fuse-e"        , OPTION_AVR_WRITE_FUSE_E     , true,
		"--target-clock"            , OPTION_TARGET_CLOCK         , false,
		"--prompt"                  , OPTION_PROMPT               , false,
		0
	};

	bool error = false;
	bool opencomm = false;

	unsigned short address = 0;

	for (int pass = 0; pass < 2; pass++) {
		if (pass == 1) {
			if (opencomm) {
				try {
					switch (_option.device.id) {
					case H8_ID(0x3664):
					case H8_ID(0x3048):
					case H8_ID(0x3052):
					case H8_ID(0x3067):
					case H8_ID(0x3068):
					case H8_ID(0x3069):
						_option.speed = 9600;
						break;
					}
					if (!OpenPort()) {
						fprintf(stderr, "Could not open device '%s'\n", _option.port.c_str());
						return 1;
					}
				} catch (Exception const &e) {
					puts(e.message.c_str());
					return 1;
				}
			}
		}

		for (int argi = 1; argi < argc; argi++) {
			char const *arg = argv[argi];
			if (arg[0] == '-') {
				bool unknown = true;
				for (int i = 0; optiontable[i].text; i++) {
					size_t n = strlen(optiontable[i].text);
					if (strncmp(arg, optiontable[i].text, n) == 0) {
						if (arg[n] == 0 || arg[n] == '=') {
							char const *val = "";
							if (optiontable[i].with_value) {
								if (arg[n] == '=') {
									val = arg + n + 1;
								} else if (argi + 1 < argc && argv[argi + 1][0] != '-') {
									argi++;
									val = argv[argi];
								}
							}
							try {
								switch (optiontable[i].id) {
								case OPTION_TEST:
									if (pass == 0) {
										opencomm = true;
									} else {
										do_test();
									}
									break;

									// PIC

								case OPTION_ERASE_PIC12F629:
									if (pass == 0) {
										opencomm = true;
									} else {
										pic_erase_pic12f629();
									}
									break;
								case OPTION_PIC_BULK_ERASE_P:
									if (pass == 0) {
										opencomm = true;
									} else {
										pic_bulk_erase_program_memory();
									}
									break;
								case OPTION_PIC_BULK_ERASE_D:
									if (pass == 0) {
										opencomm = true;
									} else {
										pic_bulk_erase_data_memory();
									}
									break;
								case OPTION_PIC_INQUIRE:
									if (pass == 0) {
										opencomm = true;
									} else {
										pic_inquire();
									}
									break;


								case OPTION_DISABLE_CODE_PROTECT:
									if (pass == 0) {
										opencomm = true;
									} else {
										pic_disable_code_protect();
									}
									break;

								case OPTION_PIC_CONFWORD:
									if (pass == 0) {
										opencomm = true;
									} else {
										if (*val) {
											unsigned short data = (unsigned short)strtoul(val, 0, 16);
											pic_write_confword(data);
										} else {
											pic_read_confword();
										}
									}
									break;

								// AVR

								case OPTION_AVR_CHIP_ERASE:
									if (pass == 0) {
										opencomm = true;
									} else {
										avr_chip_erase();
									}
									break;

								case OPTION_AVR_READ_FUSE:
									if (pass == 0) {
										opencomm = true;
									} else {
										avr_read_fuse();
									}
									break;

								case OPTION_AVR_WRITE_FUSE_L:
									if (pass == 0) {
										opencomm = true;
									} else {
										unsigned char data = (unsigned char)strtoul(val, 0, 16);
										avr_write_fuse_l(data);
									}
									break;

								case OPTION_AVR_WRITE_FUSE_H:
									if (pass == 0) {
										opencomm = true;
									} else {
										unsigned char data = (unsigned char)strtoul(val, 0, 16);
										avr_write_fuse_h(data);
									}
									break;

								case OPTION_AVR_WRITE_FUSE_E:
									if (pass == 0) {
										opencomm = true;
									} else {
										unsigned char data = (unsigned char)strtoul(val, 0, 16);
										avr_write_fuse_e(data);
									}
									break;

								//

								case OPTION_AVR_INQUIRE:
									if (pass == 0) {
										opencomm = true;
									} else {
										avr_inquire();
									}
									break;

								case OPTION_READ:
									if (pass == 0) {
										opencomm = true;
									} else {
										Read();
									}
									break;

								case OPTION_WRITE:
									if (pass == 0) {
										opencomm = true;
									} else {
										Write();
									}
									break;

								case OPTION_INPUT:
									if (pass == 0) {
										_option.input_file = val;
										//_option.input_console = false;
									}
									break;
								case OPTION_OUTPUT:
									if (pass == 0) {
										_option.output_file = val;
										//_option.output_console = false;
									}
									break;
								//case OPTION_INPUT_CONSOLE:
								//	if (pass == 0) {
								//		_option.input_console = true;
								//	}
								//	break;
								//case OPTION_OUTPUT_CONSOLE:
								//	if (pass == 0) {
								//		_option.output_console = true;
								//	}
								//	break;
								case OPTION_PORT:
									if (pass == 0) {
										std::string devpath;
										if (isdigit(*val)) {
											int n = atoi(val);
											devpath = get_port_path(n);
										} else if (*val == '/' || *val == '\\') {
											devpath = val;
										} else if (strnicmp(val, "com", 3) == 0) {
											int n = atoi(val + 3);
											if (n > 0) n--;
											devpath = get_port_path(n);
										}
										_option.port = devpath;
									}
									break;
								case OPTION_PROGMEMSIZE:
									if (pass == 0) {
										_option.device.progmemsize = (unsigned short)strtoul(val, 0, 16);
									}
									break;
								case OPTION_DATAMEMSIZE:
									if (pass == 0) {
										_option.device.datamemsize = (unsigned short)strtoul(val, 0, 16);
									}
									break;
								case OPTION_DEVICE:
									if (pass == 0) {
										DeviceInfo info;
										if (!GetDeviceInfo(val, &info)) {
											fprintf(stderr, "Unknown device name '%s'\n", val);
											error = true;
										}
										_option.device = info;
									}
									break;
								case OPTION_FORMAT:
									if (pass == 0) {
										if (stricmp(val, "mot") == 0 || stricmp(val, "srec") == 0) {
											_option.format = FORMAT_SREC;
										} else if (stricmp(val, "ihex") == 0 || stricmp(val, "inhx8m") == 0) {
											_option.format = FORMAT_INHX8M;
										} else {
											fprintf(stderr, "Unknown format '%s'\n", val);
											error = true;
										}
									}
									break;

								case OPTION_VERSION:
									if (pass == 0) {
										printf("XP, Extreme Programmer, v" VERSION "\n");
									}
									break;

								case OPTION_FIRMWARE_VERSION:
									if (pass == 0) {
										opencomm = true;
									} else {
										FirmwareVersion();
									}
									break;

								case OPTION_TARGET_CLOCK:
									if (pass == 0) {
										_option.target_clock = atol(val);
									}
									break;

								case OPTION_PROMPT:
									if (pass == 0) {
										_option.prompt = true;
									}
									break;

								default:
									fprintf(stderr, "?\n");
									error = true;
								}
							} catch (Exception const &e) {
								if (!e.message.empty()) {
									puts(e.message.c_str());
								}
								error = true;
							}
							unknown = false;
							break;
						}
					}
				}
				if (unknown) {
					fprintf(stderr, "Invalid option '%s'\n", arg);
					error = true;
				}
			}
		}
		if (pass == 1) {
			if (opencomm) {
				ClosePort();
			}
		}
		if (error) {
			return 1;
		}
	}

	return 0;
}

int main(int argc, char **argv)
{
	XP xp;
	xp.main(argc, argv);
}


